"""Upcloud cloud methods"""

import asyncio
import logging
import time
from concurrent.futures.thread import ThreadPoolExecutor
from functools import lru_cache
from typing import Optional

from asyncssh.public_key import SSHKey
from upcloud_api import CloudManager, Server, Storage, login_user_block

from ..config import ConfigCloudUpcloud
from .protocols import PCloudConfig
from .utils import get_rnd_name

executor = ThreadPoolExecutor(max_workers=5)


@lru_cache(maxsize=None)
def get_client(cfg: ConfigCloudUpcloud) -> CloudManager:
    """Get Upcloud client"""
    client = CloudManager(cfg.login, cfg.password)
    client.authenticate()
    return client


def upcloud_create_node_sync(
    log: logging.Logger,
    cfg: ConfigCloudUpcloud,
    key: SSHKey,
    cloud_config: Optional[PCloudConfig] = None,
) -> str:
    """Create node"""
    client = get_client(cfg)

    login_user = login_user_block(
        username=cfg.username,
        ssh_keys=[key.export_public_key("openssh").decode("utf-8")],
        create_password=False,
    )
    server = client.create_server(
        Server(
            core_number=8,
            memory_amount=4096,
            hostname=get_rnd_name("node"),
            zone="uk-lon1",
            storage_devices=[Storage(os="Debian 10.0", size=40)],
            login_user=login_user,
            user_data=cloud_config.render() if cloud_config else None,
        )
    )
    ip_addr = server.get_public_ip()
    log.info("CREATED %s", ip_addr)
    return ip_addr


async def upcloud_create_node(
    log: logging.Logger,
    cfg: ConfigCloudUpcloud,
    key: SSHKey,
    cloud_config: Optional[PCloudConfig] = None,
) -> str:
    """Create node"""
    return await asyncio.get_running_loop().run_in_executor(
        executor, upcloud_create_node_sync, log, cfg, key, cloud_config
    )


def upcload_delete_node_sync(
    log: logging.Logger,
    cfg: ConfigCloudUpcloud,
    host: str,
):
    """Delete node"""
    client = get_client(cfg)
    for server in client.get_servers():
        if server.get_public_ip() == host:
            server.stop()
            log.info("WAITING FOR STOP...")
            time.sleep(20)
            while True:
                try:
                    server.destroy()
                except Exception:
                    time.sleep(5)
                else:
                    break
            for storage in server.storage_devices:
                storage.destroy()
            log.info("DELETED %s", host)
            break
    else:
        log.info("NODE %s NOT DELETED AS UNKNOWN", host)


async def upcload_delete_node(
    log: logging.Logger,
    cfg: ConfigCloudUpcloud,
    host: str,
):
    """Delete node"""
    return await asyncio.get_running_loop().run_in_executor(
        executor, upcload_delete_node_sync, log, cfg, host
    )



#!/usr/bin/env python3

import asyncio
import json
import logging
import re
from pathlib import PurePath, PureWindowsPath
from typing import AsyncGenerator, Optional, Pattern, Sequence, Union

from asyncssh.connection import SSHClientConnection
from asyncssh.sftp import SFTPClient

from ..config import LocalArchiveDeploy, LocalFilesDeploy, RemoteArchiveDeploy
from .common import ProcessInfo
from .protocol import OuterRunCallable, PEngineRepository, PProcessInfo, QuoteCallable


class MyPureWindowsPath(PureWindowsPath):
    @classmethod
    def _parse_args(cls, path):
        drv, root, parts = cls._parse_args(path)  # pylint: disable=no-member
        # prevent leading slash like \C:\Users\user
        if not drv and root == "\\" and len(parts) > 2 and parts[0] == "\\":
            drv = parts[1]
            parts = parts[1:]
        # prevent eating first part when parsing PurePath instance
        if len(parts) > 1 and drv == parts[0] and not root:
            root = "\\"

        return drv, root, parts


def windows_quote(s: str) -> str:
    return "'{}'".format(str(s).replace("'", "''"))


async def windows_get_cpu_cores(run: OuterRunCallable) -> int:
    """
    Get number of CPU cores
    :raises asyncssh.Error: An SSH error has occurred.
    """
    res = await run("[environment]::ProcessorCount")
    try:
        return int(res.stdout and res.stdout.strip() or "1")
    except ValueError:
        return 1


async def windows_list_processes(
    conn: SSHClientConnection, query: Optional[str] = None
) -> AsyncGenerator[PProcessInfo, None]:
    """
    Returns information about all running processes
    :raises asyncssh.Error: An SSH error has occurred.
    """
    where_pipe_cmd = f"| ?{{ {query} }}" if query else ""
    get_process_cmd = f"Get-CimInstance Win32_Process {where_pipe_cmd}"
    inline_obj = "@{'pid' = $_.ProcessId; 'name' = $_.Name; 'command' = $_.CommandLine}"
    for_each_cmd = f"%{{ {inline_obj} | ConvertTo-Json -compress }}"
    ps_cmd = f"{get_process_cmd} | {for_each_cmd}"
    async with conn.create_process(ps_cmd) as proc:
        async for line in proc.stdout:
            try:
                data = json.loads(line)
                if not data["command"]:
                    data["command"] = data["name"]
                assert isinstance(data["pid"], int)
                assert isinstance(data["name"], str)
                assert isinstance(data["command"], str)
                # skip self
                if (
                    data["name"] == "powershell.exe"
                    and "Get-CimInstance Win32_Process" in data["command"]
                ):
                    continue
                yield ProcessInfo(**data)
            except Exception:
                continue


async def windows_pgrep(
    conn: SSHClientConnection,
    quote: QuoteCallable,
    pattern: Union[str, Pattern[str]],
    full=True,
) -> AsyncGenerator[PProcessInfo, None]:
    """
    Returns information about running processes, that name matches a pattern.
    If `full`, check match against name or full cmd.
    :raises asyncssh.Error: An SSH error has occurred.
    """
    str_pattern = pattern.pattern if isinstance(pattern, re.Pattern) else pattern
    match_tail = ["-match", quote(str_pattern)]
    name_expr = ["$_.Name", *match_tail]
    cmd_expr = ["$_.CommandLine", *match_tail]
    if full:
        where_expr = " ".join([*name_expr, "-or", *cmd_expr])
    else:
        where_expr = " ".join(name_expr)
    async for x in windows_list_processes(conn, query=where_expr):
        yield x


async def deploy_local_files(
    sftp: SFTPClient,
    engine_dir: PurePath,
    files: Sequence[PurePath],
    log: Optional[logging.Logger] = None,
):
    "Uploading binary from local; requires broadband connection"

    async def upload(src: PurePath, dst: PurePath):
        if log:
            log.debug(f"Uploading file {str(src)} to {str(dst)}")
        await sftp.put([str(src)], str(dst))

    await asyncio.gather(*map(lambda x: upload(x, engine_dir / x.name), files))


async def deploy_local_archive(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engine_dir: PurePath,
    archive: PurePath,
    log: Optional[logging.Logger] = None,
):
    """
    Upload local archive.
    Binary may be gzipped, without subfolders, with an arbitrary archive name.
    """
    rpath = engine_dir / archive.name
    if log:
        log.debug(f"Uploading {archive.name} to {str(rpath)}...")
    await sftp.put([str(archive)], engine_dir)
    if log:
        log.debug(f"Unarchiving {archive.name}...")
    await run(
        f"""Expand-Archive {quote(str(rpath))} `
            -DestinationPath {quote(str(engine_dir))} `
            -Force""",
        check=True,
    )
    await sftp.remove(rpath)


async def deploy_remote_archive(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engine_dir: PurePath,
    url: str,
    log: Optional[logging.Logger] = None,
):
    """
    Downloading binary from a trusted non-public address.
    Binary may be gzipped, without subfolders, with an arbitrary archive name.
    """
    name = "archive.zip"
    rpath = engine_dir / name
    if log:
        log.debug(f"Downloading {url} to {str(rpath)}...")
    await run(
        f"""Invoke-WebRequest -Uri {quote(url)} `
            -OutFile {quote(str(rpath))} -Force""",
        check=True,
    )
    if log:
        log.debug(f"Unarchiving {name}...")
    await run(
        f"""Expand-Archive {quote(str(rpath))} `
            -DestinationPath {quote(str(engine_dir))} `
            -Force""",
        check=True,
    )
    await sftp.remove(rpath)


async def windows_deploy_engines(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engines: PEngineRepository,
    engines_dir: PurePath,
    log: Optional[logging.Logger] = None,
) -> None:
    """
    Setup node for target engines.
    """
    for engine in engines.values():
        if log:
            log.info(f"Setup {engine.name} engine...")
        engine_dir = PureWindowsPath(
            (await sftp.realpath(engines_dir / engine.name))[1:]
        )
        # sftp.makedirs is broken for PureWindowsPath
        await sftp.makedirs(PurePath(engine_dir), exist_ok=True)
        for deployment in engine.deployable:
            if isinstance(deployment, LocalFilesDeploy):
                await deploy_local_files(sftp, engine_dir, deployment.files, log)

            if isinstance(deployment, LocalArchiveDeploy):
                await deploy_local_archive(
                    run, quote, sftp, engine_dir, deployment.file
                )

            if isinstance(deployment, RemoteArchiveDeploy):
                await deploy_remote_archive(
                    run, quote, sftp, engine_dir, deployment.url
                )
        if log:
            log.info(f"Setup of {engine.name} engine is done...")


async def windows_setup_node(
    conn: SSHClientConnection,
    run: OuterRunCallable,
    quote: QuoteCallable,
    engines: PEngineRepository,
    engines_dir: PurePath,
    log: Optional[logging.Logger] = None,
):
    "Setup generic linux node"
    async with conn.start_sftp_client() as sftp:
        await windows_deploy_engines(run, quote, sftp, engines, engines_dir, log)



#!/usr/bin/env python3

import asyncio
import logging
from collections import UserDict
from datetime import timedelta
from operator import itemgetter
from typing import Callable, MutableMapping, Optional, Sequence, Set

from attrs import define, evolve, field
from typing_extensions import Self

from .protocol import PRemoteMachine


@define
class RemoteMachineRepository(UserDict, MutableMapping[str, PRemoteMachine]):
    log: Optional[logging.Logger]
    data: MutableMapping[str, PRemoteMachine] = field(factory=dict)
    connect_in_flight: Set[str] = field(factory=set, init=False)

    def get(
        self, __key: str, default: Optional[PRemoteMachine] = None
    ) -> Optional[PRemoteMachine]:
        return self.data.get(__key, default)

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    async def disconnect_many(self, ips: Sequence[str]) -> None:
        "Disconnect from many remote machines and remove them from registry"
        if not ips:
            return
        if self.log:
            self.log.info("Disconnecting from machines: {}".format(", ".join(ips)))

        tasks = []
        for ip, machine in list(self.data.items()):
            # guard
            if machine.meta.busy:
                continue
            if ip in ips:
                tasks.append(machine.close())
                del self.data[ip]
        await asyncio.gather(*tasks, return_exceptions=True)

    async def disconnect_all(self) -> None:
        "Disconnect from all remotes"
        await self.disconnect_many(list(self.data.keys()))

    def filter(
        self,
        busy: Optional[bool] = None,
        platforms: Optional[Sequence[str]] = None,
        free_since_gt: Optional[timedelta] = None,
        reverse_sort: bool = False,
    ) -> Self:
        "Return machines filtered and sorted by `free_since`"

        checks: Sequence[Callable[[PRemoteMachine], bool]] = []
        if busy is True:
            checks.append(lambda x: x.meta.busy is True)
        if busy is False:
            checks.append(lambda x: not x.meta.busy)
        if platforms:
            checks.append(lambda x: bool(set(platforms) & set(x.platforms)))
        if free_since_gt:
            checks.append(lambda x: x.meta.is_free_longer_than(free_since_gt))

        return evolve(
            self,
            data={
                ip: m
                for ip, m in sorted(
                    self.data.items(), key=itemgetter(1), reverse=reverse_sort
                )
                if all([x(m) for x in checks])
            },
        )



"""Yascheduler client"""

import asyncio
import logging
import sys
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from pathlib import PurePath
from typing import (
    Any,
    Callable,
    Coroutine,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from attrs import asdict

from .config import Config
from .db import DB, TaskStatus
from .scheduler import Scheduler
from .variables import CONFIG_FILE

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

ReturnT_co = TypeVar("ReturnT_co", covariant=True)
ParamT = ParamSpec("ParamT")


def to_sync(
    func: Callable[ParamT, Coroutine[Any, Any, ReturnT_co]],
) -> Callable[ParamT, ReturnT_co]:
    """
    Wraps async function and run it sync in thread.
    """

    @wraps(func)
    def outer(*args: ParamT.args, **kwargs: ParamT.kwargs):
        """
        Execute the async method synchronously in sync and async runtime.
        """
        coro = func(*args, **kwargs)
        try:
            asyncio.get_running_loop()  # Triggers RuntimeError if no running event loop

            # Create a separate thread so we can block before returning
            with ThreadPoolExecutor(1) as pool:
                return pool.submit(lambda: asyncio.run(coro)).result()
        except RuntimeError:
            return asyncio.run(coro)

    return outer


class Yascheduler:
    """Yascheduler client"""

    STATUS_TO_DO = TaskStatus.TO_DO.value
    STATUS_RUNNING = TaskStatus.RUNNING.value
    STATUS_DONE = TaskStatus.DONE.value

    config: Config
    _logger: Optional[logging.Logger] = None

    def __init__(
        self,
        config_path: Union[PurePath, str] = CONFIG_FILE,
        logger: Optional[logging.Logger] = None,
    ):
        self.config = Config.from_config_parser(config_path)
        self._logger = logger

    async def queue_submit_task_async(
        self,
        label: str,
        metadata: Mapping[str, Any],
        engine_name: str,
        webhook_onsubmit=False,
    ) -> int:
        """Submit new task"""
        yac = await Scheduler.create(config=self.config, log=self._logger)
        task = await yac.create_new_task(
            label=label,
            metadata=metadata,
            engine_name=engine_name,
            webhook_onsubmit=webhook_onsubmit,
        )
        await yac.stop()
        return task.task_id

    def queue_submit_task(
        self,
        label: str,
        metadata: Mapping[str, Any],
        engine_name: str,
        webhook_onsubmit=False,
    ) -> int:
        """Submit new task"""
        fn = to_sync(self.queue_submit_task_async)
        return fn(label, metadata, engine_name, webhook_onsubmit)

    async def queue_get_tasks_async(
        self,
        jobs: Optional[Sequence[int]] = None,
        status: Optional[Sequence[int]] = None,
    ) -> Sequence[Mapping[str, Any]]:
        """Get tasks by ids or statuses"""
        if jobs is not None and status is not None:
            raise ValueError("jobs can be selected only by status or by task ids")
        # raise ValueError if unknown task status
        status = [TaskStatus(x) for x in status] if status else None
        db = await DB.create(self.config.db)
        if status:
            tasks = await db.get_tasks_by_status(status)
        elif jobs:
            tasks = await db.get_tasks_by_jobs(jobs)
        else:
            return []
        return [asdict(t) for t in tasks]

    def queue_get_tasks(
        self,
        jobs: Optional[Sequence[int]] = None,
        status: Optional[Sequence[int]] = None,
    ) -> Sequence[Mapping[str, Any]]:
        """Get tasks by ids or statuses"""
        return to_sync(self.queue_get_tasks_async)(jobs, status)

    async def queue_get_task_async(self, task_id: int) -> Optional[Mapping[str, Any]]:
        """Get task by id"""
        for task_dict in await self.queue_get_tasks_async(jobs=[task_id]):
            return task_dict

    def queue_get_task(self, task_id: int) -> Optional[Mapping[str, Any]]:
        """Get task by id"""
        for task_dict in self.queue_get_tasks(jobs=[task_id]):
            return task_dict



#!/usr/bin/env python3
"""Local configuration"""

from configparser import SectionProxy
from pathlib import Path, PurePath
from typing import Optional, Sequence

from attrs import define, field, fields, validators

from .utils import _make_default_field, warn_unknown_fields


@define(frozen=True)
class ConfigLocal:
    """Local configuration"""

    data_dir: Path = _make_default_field(Path("./data"))
    tasks_dir: Path = _make_default_field(Path("./data/tasks"))
    engines_dir: Path = _make_default_field(Path("./data/engines"))
    keys_dir: Path = _make_default_field(Path("./data/keys"))
    webhook_url: Optional[str] = field(default=None)
    webhook_reqs_limit: int = _make_default_field(
        5, extra_validators=[validators.ge(1)]
    )
    conn_machine_limit: int = _make_default_field(
        10, extra_validators=[validators.ge(1)]
    )
    conn_machine_pending: int = _make_default_field(
        10, extra_validators=[validators.ge(1)]
    )
    allocate_limit: int = _make_default_field(20, extra_validators=[validators.ge(1)])
    allocate_pending: int = _make_default_field(1, extra_validators=[validators.ge(1)])
    consume_limit: int = _make_default_field(20, extra_validators=[validators.ge(1)])
    consume_pending: int = _make_default_field(1, extra_validators=[validators.ge(1)])
    deallocate_limit: int = _make_default_field(5, extra_validators=[validators.ge(1)])
    deallocate_pending: int = _make_default_field(
        1, extra_validators=[validators.ge(1)]
    )

    def get_private_keys(self) -> Sequence[PurePath]:
        "List private key file paths"
        filepaths = filter(lambda x: x.is_file(), Path(self.keys_dir).iterdir())
        return list(filepaths)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        return [f.name for f in fields(cls)]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigLocal":
        "Create config from config parser's section"

        warn_unknown_fields(cls.get_valid_config_parser_fields(), sec)

        data_dir = Path(sec.get("data_dir", "./data")).resolve()
        return ConfigLocal(
            data_dir,
            tasks_dir=Path(sec.get("tasks_dir", str(data_dir / "tasks"))).resolve(),
            engines_dir=Path(
                sec.get("engines_dir", str(data_dir / "engines"))
            ).resolve(),
            keys_dir=Path(sec.get("keys_dir", str(data_dir / "keys"))).resolve(),
            webhook_reqs_limit=sec.getint("webhook_reqs_limit"),
            webhook_url=sec.get("webhook_url"),
            conn_machine_limit=sec.getint("conn_machine_limit"),
            conn_machine_pending=sec.getint("conn_machine_pending"),
            allocate_limit=sec.getint("allocate_limit"),
            allocate_pending=sec.getint("allocate_pending"),
            consume_limit=sec.getint("consume_limit"),
            consume_pending=sec.getint("consume_pending"),
            deallocate_limit=sec.getint("deallocate_limit"),
            deallocate_pending=sec.getint("deallocate_pending"),
        )



#!/usr/bin/env python3
"""Repository for Engines"""

import json
from collections import UserDict
from configparser import ConfigParser
from itertools import chain
from pathlib import PurePath
from typing import Any, Callable, Mapping, Sequence

from attrs import Attribute, asdict, define, field, validators
from typing_extensions import Self

from .engine import Engine


def _value_serializer(_: type, __: Attribute, value: Any) -> Any:
    "Serialize PurePath as string"
    if isinstance(value, PurePath):
        return str(value)
    return value


@define
class EngineRepository(UserDict, Mapping[str, Engine]):
    """Repository of Engines"""

    engines_dir: PurePath = field()
    data: Mapping[str, Engine] = field(
        factory=dict,
        validator=[
            validators.deep_mapping(
                key_validator=validators.instance_of(str),
                value_validator=validators.instance_of(Engine),
            )
        ],
    )

    def __setitem__(self, _: str, __: Engine):
        raise NotImplementedError()

    def __delitem__(self, _: str):
        raise NotImplementedError()

    def __hash__(self) -> int:
        return hash(
            json.dumps(asdict(self, value_serializer=_value_serializer), sort_keys=True)
        )

    def get(self, key, default=None):
        if default:
            return self.data.get(key, default)
        return self.data.get(key)

    def values(self):
        return self.data.values()

    def filter(self, filter_func: Callable[[Engine], bool]) -> Self:
        "Filter Engines by callable and return new Repository"
        new_data = dict(filter(lambda x: filter_func(x[1]), self.data.items()))
        return self.__class__(engines_dir=self.engines_dir, data=new_data)

    def filter_platforms(self, platforms: Sequence[str]) -> Self:
        "Filter Engines by platforms and return new Repository"
        return self.filter(lambda x: bool(set(x.platforms) & set(platforms)))

    def get_platform_packages(self) -> Sequence[str]:
        "Collect all platform pacakges from engines"
        mapped = map(lambda x: x.platform_packages, self.values())
        return list(set(chain(*mapped)))

    @classmethod
    def from_config_parser(cls, cfg: ConfigParser, engines_dir: PurePath) -> Self:
        "Create config from path or config file contents"
        snames = filter(lambda x: x.startswith("engine."), cfg.sections())
        data = {}
        for sname in snames:
            engine = Engine.from_config_parser_section(cfg[sname], engines_dir)
            data[engine.name] = engine
        return cls(engines_dir=engines_dir, data=data)



#!/usr/bin/env python3
"""Main config module"""

from configparser import ConfigParser
from pathlib import PurePath
from typing import Sequence, Union

from attrs import define, field, validators

from .cloud import ConfigCloud, ConfigCloudAzure, ConfigCloudHetzner, ConfigCloudUpcloud
from .db import ConfigDb
from .engine_repository import EngineRepository
from .local import ConfigLocal
from .remote import ConfigRemote


@define(frozen=True)
class Config:
    """Main config module"""

    db: ConfigDb = field(validator=[validators.instance_of(ConfigDb)])
    local: ConfigLocal = field(validator=[validators.instance_of(ConfigLocal)])
    remote: ConfigRemote = field(validator=[validators.instance_of(ConfigRemote)])
    clouds: Sequence[ConfigCloud]
    engines: EngineRepository = field(
        validator=[validators.instance_of(EngineRepository)]
    )

    @classmethod
    def from_config_parser(cls, files: Union[str, bytes, PurePath]) -> "Config":
        "Create Config from path or config file contents"
        config = ConfigParser()
        config.read(files)

        for sec_name in ["db", "local", "remote", "clouds"]:
            if not config.has_section(sec_name):
                config.add_section(sec_name)

        local = ConfigLocal.from_config_parser_section(config["local"])
        remote = ConfigRemote.from_config_parser_section(config["remote"])

        # config prefixes
        cloud_prefixes = set(map(lambda x: x.split("_")[0], config.options("clouds")))
        # inherit username
        for prefix in cloud_prefixes:
            key = f"{prefix}_user"
            if key not in config.options("clouds"):
                config["clouds"][key] = remote.username
        # available cloud config models
        cloud_variants = (
            ConfigCloudAzure,
            ConfigCloudHetzner,
            ConfigCloudUpcloud,
        )
        # intersection
        cloud_variants_match = filter(
            lambda x: x.prefix in cloud_prefixes, cloud_variants
        )
        # instantiate
        clouds = map(
            lambda x: x.from_config_parser_section(config["clouds"]),
            cloud_variants_match,
        )

        return cls(
            db=ConfigDb.from_config_parser_section(config["db"]),
            local=local,
            remote=remote,
            clouds=list(clouds),
            engines=EngineRepository.from_config_parser(config, local.engines_dir),
        )



#!/usr/bin/env python3
"""Configuration module"""

from .cloud import ConfigCloud, ConfigCloudAzure, ConfigCloudHetzner, ConfigCloudUpcloud
from .config import Config
from .db import ConfigDb
from .engine import (
    Deploy,
    Engine,
    LocalArchiveDeploy,
    LocalFilesDeploy,
    RemoteArchiveDeploy,
)
from .engine_repository import EngineRepository
from .local import ConfigLocal
from .remote import ConfigRemote

__all__ = [
    "Config",
    "ConfigCloud",
    "ConfigCloudAzure",
    "ConfigCloudHetzner",
    "ConfigCloudUpcloud",
    "ConfigDb",
    "ConfigLocal",
    "ConfigRemote",
    "Deploy",
    "Engine",
    "EngineRepository",
    "LocalArchiveDeploy",
    "LocalFilesDeploy",
    "RemoteArchiveDeploy",
]



#!/usr/bin/env python

import asyncio
import logging
from asyncio.locks import Event, Semaphore
from collections import Counter
from datetime import datetime, timedelta
from functools import partial
from pathlib import Path, PurePath, PurePosixPath
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import aiohttp
import asyncssh
import backoff
from asyncssh.sftp import SFTPClient, SFTPError
from attrs import asdict, define, evolve, field
from typing_extensions import Self

from .clouds import CloudAPIManager, PCloudAPIManager
from .config import Config, Engine
from .db import DB, NodeModel, TaskModel, TaskStatus
from .queue import TUMsgId, TUMsgPayload, UMessage, UniqueQueue
from .remote_machine import (
    AllSSHRetryExc,
    PRemoteMachine,
    RemoteMachine,
    RemoteMachineRepository,
    SFTPRetryExc,
)
from .time import asleep_until
from .variables import CONFIG_FILE

logging.basicConfig(level=logging.INFO)


def get_logger(log_file, level: int = logging.INFO):
    logging.captureWarnings(True)
    logger = logging.getLogger("yascheduler")
    logger.setLevel(level)

    backoff_logger = logging.getLogger("backoff")
    backoff_logger.setLevel(logging.ERROR if level >= logging.INFO else logging.DEBUG)

    if log_file:
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        backoff_logger.addHandler(fh)

    return logger


@define(frozen=True)
class WebhookPayload:
    task_id: int = field()
    status: int = field()
    custom_params: Mapping[str, Any] = field(factory=dict)


@define
class Scheduler:
    config: Config = field()
    db: DB = field()
    clouds: PCloudAPIManager = field()
    log: logging.Logger = field()
    bg_jobs: Set[asyncio.Task] = field(factory=set, init=False)
    conn_machine_q: UniqueQueue[str, NodeModel] = field(init=False)
    allocate_q: UniqueQueue[int, TaskModel] = field(init=False)
    consume_q: UniqueQueue[int, TaskModel] = field(init=False)
    deallocate_q: UniqueQueue[str, NodeModel] = field(init=False)
    cancellation_event: Event = field(factory=Event, init=False)
    remote_machines: RemoteMachineRepository = field()
    sleep_interval: int = field(default=1)
    http: aiohttp.ClientSession = field(factory=aiohttp.ClientSession, init=False)
    webhook_sem: Semaphore = field(init=False)

    def __attrs_post_init__(self):
        lcfg = self.config.local
        self.webhook_sem = Semaphore(lcfg.webhook_reqs_limit)
        self.conn_machine_q = UniqueQueue(
            "conn_machine", maxsize=lcfg.conn_machine_pending
        )
        self.allocate_q = UniqueQueue("allocate", maxsize=lcfg.allocate_pending)
        self.consume_q = UniqueQueue("consume", maxsize=lcfg.consume_pending)
        self.deallocate_q = UniqueQueue("deallocate", maxsize=lcfg.deallocate_pending)

    @classmethod
    async def create(
        cls,
        config: Optional[Config] = None,
        log: Optional[logging.Logger] = None,
    ) -> Self:
        "Async object initialization"
        if log:
            log = log.getChild(cls.__name__)
        else:
            log = logging.getLogger(cls.__name__)
        cfg = config or Config.from_config_parser(CONFIG_FILE)
        db = await DB.create(cfg.db)
        clouds = await CloudAPIManager.create(
            db=db,
            local_config=cfg.local,
            remote_config=cfg.remote,
            cloud_configs=cfg.clouds,
            engines=cfg.engines,
            log=log,
        )

        return cls(
            config=cfg,
            db=db,
            clouds=clouds,
            log=log,
            remote_machines=RemoteMachineRepository(log=log),
            sleep_interval=min(x.sleep_interval for x in cfg.engines.values()),
        )

    async def clouds_get_capacity(self) -> int:
        "Get capacity of all clouds"
        ccap = await self.clouds.get_capacity()
        n_busy_cloud_nodes = sum(x.current for x in ccap.values())
        max_nodes = sum(x.config.max_nodes for x in self.clouds.apis.values())
        diff = max_nodes - n_busy_cloud_nodes
        return max(0, diff)

    async def do_task_webhook(
        self, task_id: int, metadata: Mapping[str, Any], status: TaskStatus
    ):
        "Send webhook with task status"
        retry = backoff.on_exception(backoff.fibo, aiohttp.ClientError, max_time=60)
        url = metadata.get("webhook_url")
        if not url:
            return
        async with self.webhook_sem:
            self.log.info(f"Executing webhook of type {status.value} to {url}")
            payload = WebhookPayload(
                task_id, status.value, metadata.get("webhook_custom_params", {})
            )
            try:
                async with retry(self.http.post)(url, data=asdict(payload)) as resp:
                    if resp.ok:
                        return
                    self.log.warn(
                        "Webhook for task_id=%s bad response: %s %s",
                        task_id,
                        resp.status,
                        resp.reason,
                    )
                    if self.log.isEnabledFor(logging.DEBUG):
                        self.log.debug(
                            "Webhook for task_id=%s response: %s",
                            task_id,
                            (await resp.text("utf-8")),
                        )
            except Exception as err:
                self.log.error("Webhook for task_id=%s failed: %s", task_id, err)

    async def create_new_task(
        self,
        label: str,
        metadata: Mapping[str, Any],
        engine_name: str,
        webhook_onsubmit: bool = False,
    ) -> TaskModel:
        "Create new task in DB"
        if engine_name not in self.config.engines:
            raise RuntimeError("Engine %s requested, but not supported" % engine_name)

        for input_file in self.config.engines[engine_name].input_files:
            if input_file not in metadata:
                raise RuntimeError("Input file %s was not provided" % input_file)

        meta_add = [("engine", engine_name)]
        new_meta = dict(list(metadata.items()) + meta_add)

        task = await self.db.add_task(
            label, ip_addr=None, status=TaskStatus.TO_DO, metadata=new_meta
        )
        dt_str = datetime.now().strftime("%Y%m%d_%H%M%S")
        remote_folder = self.config.remote.tasks_dir / "{}_{}".format(
            dt_str, task.task_id
        )
        new_meta.update({"remote_folder": str(remote_folder)})
        await self.db.update_task_meta(task.task_id, new_meta)
        await self.db.commit()
        if webhook_onsubmit:
            await self.do_task_webhook(task.task_id, new_meta, TaskStatus.TO_DO)
        self.log.info(":::submitted: %s" % label)
        return task

    async def upload_task_data(
        self,
        sftp: SFTPClient,
        task: TaskModel,
        remote_dir: PurePath,
        input_files: Sequence[str],
    ) -> bool:
        "Upload task data to remote machine"
        try:
            await sftp.makedirs(PurePosixPath(remote_dir), exist_ok=True)
        except asyncssh.misc.Error as err:
            self.log.error(
                "Create %s - SFTPError: %s (%s) (task_id=%s)"
                % (str(remote_dir), err.reason, err.code, task.task_id)
            )
            raise err
        for input_file in input_files:
            r_input_file = remote_dir / input_file
            try:
                async with sftp.open(r_input_file.as_posix(), pflags_or_mode="w") as f:
                    await f.write(task.metadata[input_file])
            except asyncssh.misc.Error as err:
                self.log.error(
                    "Write %s - SFTPError: %s (%s)"
                    % (str(r_input_file), err.reason, err.code)
                )
                raise err
        return True

    async def start_task_on_machine(
        self,
        machine: PRemoteMachine,
        engine: Engine,
        task: TaskModel,
    ) -> bool:
        "Run task on remote machine"
        self.log.info(
            "Submitting task_id=%s %s with %s to %s"
            % (task.task_id, task.label, engine.name, machine.hostname)
        )
        assert task.metadata.get("remote_folder")
        machine.meta.busy = True
        remote_folder = machine.path(task.metadata["remote_folder"])

        async with machine.sftp() as sftp:
            try:
                root_dir = machine.path(await sftp.realpath("."))
                task_dir = (
                    remote_folder
                    if remote_folder.is_absolute()
                    else root_dir / remote_folder
                )
                if self.config.remote.engines_dir.is_absolute():
                    engine_path = self.config.remote.engines_dir / engine.name
                else:
                    engine_path = (
                        root_dir / self.config.remote.engines_dir / engine.name
                    )
                # upload files
                await self.upload_task_data(sftp, task, task_dir, engine.input_files)
            except Exception as err:
                self.log.error(f"Can't upload task_id={task.task_id} files: {err}")
                raise err

        # start task
        try:
            # detect cpus
            node = await self.db.get_node(task.ip)
            ncpus = node and node.ncpus or await machine.get_cpu_cores()
            # placeholders {task_path}, {engine_path} and {ncpus} are supported
            run_cmd = engine.spawn.format(
                engine_path=str(engine_path),
                task_path=machine.quote(str(task_dir)),
                ncpus=ncpus,
            )
            await machine.run_bg(run_cmd, cwd=str(task_dir))
        except Exception as err:
            self.log.error("SSH spawn cmd error: %s" % err)
            raise err

        return True

    async def allocate_task(self, task: TaskModel) -> bool:
        "Allocate task to a free remote machine or ask allocation of new cloud machine"
        self.log.debug(f"Allocating task {task.task_id}")
        engine_name: Optional[str] = task.metadata.get("engine", None)
        engine: Optional[Engine] = self.config.engines.get(engine_name)
        if engine is None:
            self.log.warning(
                "Unsupported engine '{}' for task_id={}".format(
                    engine_name, task.task_id
                )
            )
            await self.db.set_task_error(
                task.task_id, metadata=task.metadata, error="unsupported engine"
            )
            await self.do_task_webhook(task.task_id, task.metadata, TaskStatus.DONE)
            return False

        busy_node_ips = [
            t.ip for t in await self.db.get_tasks_by_status((TaskStatus.RUNNING,))
        ]
        free_machines = {
            ip: m
            for ip, m in self.remote_machines.filter(
                busy=False, platforms=engine.platforms, reverse_sort=True
            ).items()
            if ip not in busy_node_ips
        }
        if free_machines:
            self.log.debug(
                "Free machines with platform match: %s"
                % ", ".join(free_machines.keys())
            )
        for ip, machine in free_machines.items():
            task_m = evolve(task, ip=ip)
            self.log.debug(f"Allocate task {task.task_id} to machine {ip}")
            if await self.start_task_on_machine(machine, engine, task_m):
                self.log.debug(f"Task {task.task_id} allocated to machine {ip}")
                await machine.start_occupancy_check(engine)
                await self.db.set_task_running(task.task_id, task_m.ip)
                await self.db.commit()
                await self.do_task_webhook(
                    task.task_id, task_m.metadata, TaskStatus.RUNNING
                )
                self.clouds.mark_task_done(task.task_id)
                return True

        # free machine not found - try to allocate new node
        self.log.debug(
            f"No free machine for task {task.task_id} - want to allocate new node"
        )
        await self.clouds.allocate(
            task.task_id, want_platforms=engine.platforms, throttle=True
        )
        return False

    async def consume_task(self, machine: PRemoteMachine, task: TaskModel):
        "Consume done tasks"
        meta = task.metadata
        local_folder: Union[str, None] = meta.get("local_folder")
        remote_folder: str = meta["remote_folder"]
        engine: Engine = self.config.engines[meta["engine"]]
        # NOTE: PureWindowsPath is not supported but posix is fine
        output_files = [
            str(PurePosixPath(remote_folder) / x) for x in engine.output_files
        ]
        if local_folder:
            store_folder = Path(local_folder)
        else:
            store_folder = self.config.local.tasks_dir / Path(remote_folder).name
        await asyncio.get_running_loop().run_in_executor(
            None, store_folder.mkdir, 0o777, True, True
        )

        meta_add = [
            ("remote_folder", remote_folder),
            ("local_folder", str(store_folder)),
        ]

        sftp_errors: List[Tuple[Optional[str], Exception]] = []
        sftp_get_retry = backoff.on_exception(backoff.fibo, SFTPRetryExc, max_time=60)

        async def job():
            async with machine.sftp() as sftp:
                for out_file in output_files:
                    try:
                        await sftp_get_retry(sftp.get)(
                            out_file, store_folder, preserve=True
                        )
                    except (OSError, SFTPError) as err:
                        sftp_errors.append((out_file, err))
                        self.log.warning(
                            "Cannot download file for task_id=%s from %s: %s",
                            task.task_id,
                            out_file,
                            err,
                        )
                await sftp.rmtree(
                    machine.path(remote_folder)
                )  # comment to keep the raw files at the working node (not recommended)

        try:
            await sftp_get_retry(job)()
        except Exception as err:
            self.log.warning(f"Cannot scp from {remote_folder}: {err}")
            sftp_errors.append((remote_folder, err))

        if sftp_errors:
            meta_add.append(("error", {p: str(e) for p, e in sftp_errors}))

        new_meta = dict(list(task.metadata.items()) + meta_add)
        if "error" in new_meta:
            await self.db.set_task_error(task.task_id, new_meta)
            await self.do_task_webhook(task.task_id, new_meta, TaskStatus.DONE)
        else:
            await self.db.set_task_done(task.task_id, new_meta)
            await self.do_task_webhook(task.task_id, new_meta, TaskStatus.DONE)
        await self.db.commit()
        self.log.info(
            "task_id=%s %s done and saved in %s"
            % (task.task_id, task.label, store_folder)
        )
        self.clouds.mark_task_done(task.task_id)

    #
    # Background workers
    #

    async def print_stats(self):
        "Print usage statistics to the log"
        while not self.cancellation_event.is_set():
            end_time = datetime.now() + timedelta(seconds=10)
            ncounters = await self.db.count_nodes_by_status()
            tcounters = await self.db.count_tasks_by_status()
            tmpl = (
                "THREADS: {tasks} "
                "NODES: busy:{n_busy}/enabled:{n_enabled}/total:{n_total} "
                "TASKS: run:{t_run}/todo:{t_todo}/done:{t_done}"
            )
            msg = tmpl.format(
                tasks=len(asyncio.all_tasks()),
                n_busy=len(self.remote_machines.filter(busy=True).keys()),
                n_enabled=ncounters[True],
                n_total=sum(ncounters.values()),
                t_run=tcounters[TaskStatus.RUNNING],
                t_todo=tcounters[TaskStatus.TO_DO],
                t_done=tcounters[TaskStatus.DONE],
            )
            self.log.info(msg)

            queues = [
                self.conn_machine_q,
                self.allocate_q,
                self.deallocate_q,
                self.consume_q,
            ]
            qmsgs = [f"{q.name}: {q.psize()}/{q.qsize()}" for q in queues]
            self.log.info("QUEUES: %s" % " ".join(qmsgs))
            await asleep_until(end_time)

    async def connect_machine_producer(
        self,
    ) -> AsyncGenerator[UMessage[str, NodeModel], None]:
        """Produce messages with new machines for connecting"""
        enabled_nodes = await self.db.get_enabled_nodes()
        new_nodes = [
            n for n in enabled_nodes if n.ip not in self.remote_machines.keys()
        ]
        for node in new_nodes:
            yield UMessage(node.ip, node)

    async def connect_machine_consumer(self, msg: UMessage[str, NodeModel]):
        """Connect to machines"""
        node = msg.payload
        keys = await asyncio.get_running_loop().run_in_executor(
            None, self.config.local.get_private_keys
        )

        jump_host = self.config.remote.jump_host
        jump_username = self.config.remote.jump_username
        for cloud in self.config.clouds:
            if cloud.prefix == node.cloud:
                if cloud.jump_host and cloud.jump_username:
                    jump_host, jump_username = cloud.jump_host, cloud.jump_username

        try:
            self.remote_machines[node.ip] = await RemoteMachine.create(
                host=node.ip,
                username=node.username,
                client_keys=keys,
                logger=self.log,
                data_dir=self.config.remote.data_dir,
                engines_dir=self.config.remote.engines_dir,
                tasks_dir=self.config.remote.tasks_dir,
                connect_timeout=10,
                jump_username=jump_username,
                jump_host=jump_host,
            )
        except asyncssh.misc.Error as err:
            self.log.error(f"Can't connect to machine with error: {err}")
        except Exception as err:
            self.log.error(f"An error occuried on remote machine creation: {err}")

    async def allocator_producer(
        self,
    ) -> AsyncGenerator[UMessage[int, TaskModel], None]:
        """Produce messages with todo tasks to run"""
        ccap = await self.clouds_get_capacity()
        tlim = max(ccap, len(self.remote_machines.filter(busy=False)), 10)
        tasks = await self.db.get_tasks_by_status((TaskStatus.TO_DO,), tlim)
        if tasks:
            ids = [str(t.task_id) for t in tasks]
            self.log.debug("Want to allocate tasks: %s" % ", ".join(ids))
        for task in tasks:
            yield UMessage(task.task_id, task)

    @backoff.on_exception(backoff.fibo, AllSSHRetryExc, max_time=60)
    async def allocator_consumer(self, msg: UMessage[int, TaskModel]):
        """Allocate task to node or allocate new node"""
        await self.allocate_task(msg.payload)

    async def task_consumer_producer(
        self,
    ) -> AsyncGenerator[UMessage[int, TaskModel], None]:
        """Produce messages with running tasks for consuming"""
        tasks = await self.db.get_tasks_by_status((TaskStatus.RUNNING,))
        for task in tasks:
            yield UMessage(task.task_id, task)

    async def task_consumer_consumer(
        self, msg: UMessage[int, TaskModel], machine_not_found: Counter
    ):
        """Consume running task if done, mark failed if machine is gone"""
        broken_tasks_passes = 20
        task_id, task = msg.id, msg.payload
        machine = self.remote_machines.get(task.ip)
        if machine is None:
            self.log.warning(f"Task {task_id} - machine {task.ip} is gone")
            machine_not_found.update([task_id])
            if machine_not_found[task_id] > broken_tasks_passes:
                await self.db.set_task_error(
                    task_id, metadata=task.metadata, error="node is gone"
                )
                await self.do_task_webhook(task_id, task.metadata, TaskStatus.DONE)
            return
        # if machine state is unknown
        if machine.meta.busy is None:
            engine = self.config.engines.get(task.metadata["engine"])
            if engine:
                await machine.start_occupancy_check(engine)
        # consume
        if machine.meta.busy is False:
            self.log.debug(f"machine {machine.hostname} is free of task {task_id}")
            await self.consume_task(machine, task)

    async def deallocator_producer(
        self,
    ) -> AsyncGenerator[UMessage[str, NodeModel], None]:
        """Produce messages with nodes for deallocation"""

        # (I) disable idle nodes without linked running tasks
        tasks = await self.db.get_tasks_by_status((TaskStatus.RUNNING,))
        busy_ips = [t.ip for t in tasks]
        all_enabled_nodes = {
            n.ip: n for n in await self.db.get_enabled_nodes() if n.ip not in busy_ips
        }
        for ccfg in self.config.clouds:
            tdlim = timedelta(seconds=ccfg.idle_tolerance)
            idlers = self.remote_machines.filter(
                busy=False, reverse_sort=False, free_since_gt=tdlim
            )
            nodes_to_disable = [
                ip
                for ip, node in all_enabled_nodes.items()
                if node.cloud == ccfg.prefix and ip in idlers.keys()
            ]
            for ip in nodes_to_disable:
                await self.db.disable_node(ip)
                await self.db.commit()

        # (II) deallocate all disabled nodes without linked running tasks
        free_disabled_nodes = [
            node
            for node in await self.db.get_disabled_nodes()
            if node.ip not in busy_ips and "." in node.ip
        ]
        to_disconnect = [
            n.ip for n in free_disabled_nodes if n.ip in self.remote_machines.keys()
        ]
        await self.remote_machines.disconnect_many(to_disconnect)
        for node in free_disabled_nodes:
            yield UMessage(node.ip, node)

    async def deallocator_consumer(self, msg: UMessage[str, NodeModel]):
        """Consume running task if done, mark failed if machine is gone"""
        node = msg.payload
        if not node.cloud:
            return
        await self.clouds.deallocate(node.ip)

    async def create_producer_consumers(
        self,
        queue: UniqueQueue[TUMsgId, TUMsgPayload],
        producer: Callable[[], AsyncGenerator[UMessage[TUMsgId, TUMsgPayload], None]],
        consumer: Callable[[UMessage[TUMsgId, TUMsgPayload]], Awaitable],
        workers_num: int = 1,
    ) -> None:
        async def worker():
            while not self.cancellation_event.is_set():
                msg = await queue.get()
                try:
                    await consumer(msg)
                finally:
                    queue.item_done(msg)

        workers: Set[asyncio.Task] = set()
        [workers.add(asyncio.create_task(worker())) for _ in range(0, workers_num)]

        try:
            while not self.cancellation_event.is_set():
                end_time = datetime.now() + timedelta(seconds=self.sleep_interval)
                try:
                    async for msg in producer():
                        await queue.put(msg)
                finally:
                    await asleep_until(end_time)

        except asyncio.CancelledError:
            if not queue.empty():
                self.log.info(f"Queue {queue.name} has {queue.qsize()} items - waiting")
                await queue.join()
            [task.cancel() for task in workers]
            await asyncio.gather(*workers, return_exceptions=True)

    #
    # Lifecycle
    #

    async def start(self):
        self.log.debug(
            "Available computing engines: %s" % ", ".join(self.config.engines.keys())
        )

        self.bg_jobs.add(asyncio.create_task(self.print_stats()))

        conn_machine_co = self.create_producer_consumers(
            queue=self.conn_machine_q,
            producer=self.connect_machine_producer,
            consumer=self.connect_machine_consumer,
            workers_num=self.config.local.conn_machine_limit,
        )
        self.bg_jobs.add(asyncio.create_task(conn_machine_co))

        # wait some connected machines before allocation
        async def wait_some_machines():
            while not len(self.remote_machines):
                await asyncio.sleep(1)

        await asyncio.wait(
            [asyncio.create_task(x) for x in [wait_some_machines(), asyncio.sleep(30)]],
            return_when="FIRST_COMPLETED",
        )

        allocate_co = self.create_producer_consumers(
            queue=self.allocate_q,
            producer=self.allocator_producer,
            consumer=self.allocator_consumer,
            workers_num=self.config.local.allocate_limit,
        )
        self.bg_jobs.add(asyncio.create_task(allocate_co))

        machine_not_found = Counter()
        consume_co = self.create_producer_consumers(
            queue=self.consume_q,
            producer=self.task_consumer_producer,
            consumer=partial(
                self.task_consumer_consumer, machine_not_found=machine_not_found
            ),
            workers_num=self.config.local.consume_limit,
        )
        self.bg_jobs.add(asyncio.create_task(consume_co))

        deallocate_co = self.create_producer_consumers(
            queue=self.deallocate_q,
            producer=self.deallocator_producer,
            consumer=self.deallocator_consumer,
            workers_num=self.config.local.deallocate_limit,
        )
        self.bg_jobs.add(asyncio.create_task(deallocate_co))

        await asyncio.gather(*self.bg_jobs, return_exceptions=True)
        await asyncio.sleep(1)  # workaround aiohttp's Unclosed client session

    async def stop(self):
        self.log.info("Stopping...")
        self.cancellation_event.set()

        for task in self.bg_jobs:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        await self.clouds.stop()
        await self.remote_machines.disconnect_all()
        await self.http.close()



"""Time utils"""

import asyncio
from datetime import datetime
from time import sleep


def sleep_until(end: datetime) -> None:
    "Sleep until :end:"
    now = datetime.now()
    if now >= end:
        return
    sleep((end - now).total_seconds())


async def asleep_until(end: datetime) -> None:
    "Sleep until :end:"
    now = datetime.now()
    if now >= end:
        return
    await asyncio.sleep((end - now).total_seconds())



"""Clouds helper utilities"""

import random
import string
from pathlib import PurePath
from typing import TypeVar

from asyncssh.public_key import SSHKey

T = TypeVar("T")


def get_rnd_name(prefix: str) -> str:
    """Create random string with prefix"""
    return (
        prefix
        + "-"
        + "".join([random.choice(string.ascii_lowercase) for _ in range(8)])
    )


def get_key_name(key: SSHKey) -> str:
    """Get SSHKey's name"""
    fname_opt = key.get_filename()
    key_filename = fname_opt.decode("utf-8") if fname_opt else None
    if key_filename:
        key_filename = PurePath(key_filename).name
    key_fingerprint = key.get_fingerprint("md5").split(":", maxsplit=1)[1]
    return key_filename or key.get_comment() or key_fingerprint



"""Database utils"""

import asyncio
import json
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from enum import Enum, unique
from typing import Any, List, Mapping, Optional, Sequence, cast

import backoff
from attrs import asdict, define, field
from pg8000.native import Connection, InterfaceError
from typing_extensions import Self

from .config import ConfigDb


@unique
class TaskStatus(int, Enum):
    """Task possible states enum"""

    TO_DO = 0
    RUNNING = 1
    DONE = 2


@define(frozen=True, hash=True)
class NodeModel:
    """Node model"""

    ip: str = field()
    ncpus: Optional[int] = field()
    enabled: bool = field(default=True)
    cloud: Optional[str] = field(default=None)
    username: str = field(default="root")


@define(frozen=True)
class TaskModel:
    """Task model"""

    task_id: int = field()
    label: str = field()
    ip: str = field()
    status: TaskStatus = field(converter=TaskStatus)
    metadata: Mapping[str, Any] = field(factory=dict)
    cloud: Optional[str] = field(default=None)

    def __hash__(self) -> int:
        return hash(json.dumps(asdict(self), sort_keys=True))


@define(frozen=True)
class DB:
    """Database abstraction"""

    loop: asyncio.AbstractEventLoop = field()
    executor: ThreadPoolExecutor = field()
    conn: Connection = field()

    @staticmethod
    def create_connection(config: ConfigDb) -> Connection:
        """Create database connection"""
        return Connection(
            user=config.user,
            host=config.host,
            database=config.database,
            port=config.port,
            password=config.password,
        )

    @classmethod
    async def create(cls, config: ConfigDb, automigrate=True) -> Self:
        """Async init"""
        loop = asyncio.get_running_loop()
        exe = ThreadPoolExecutor(max_workers=1)  # pg8000 is not thread safe
        conn = await loop.run_in_executor(exe, cls.create_connection, config)
        ins = cls(loop=loop, executor=exe, conn=conn)
        if automigrate:
            await ins.migrate()
        return ins

    async def run(self, sql: str, **params):
        """Run query async with backoff"""

        @backoff.on_exception(backoff.fibo, InterfaceError, max_time=60)
        def run_fn():
            return self.conn.run(sql, **params)

        return await self.loop.run_in_executor(self.executor, run_fn)

    async def migrate(self) -> None:
        """Migrate database scheme"""
        await self.run(
            """ALTER TABLE yascheduler_nodes
            ADD COLUMN IF NOT EXISTS username VARCHAR(255) DEFAULT 'root';"""
        )

    async def commit(self):
        """Commit"""
        await self.run("COMMIT;")

    async def close(self):
        """Close connection"""
        await self.loop.run_in_executor(self.executor, self.conn.close)
        self.executor.shutdown()

    async def has_node(self, ip_addr: str) -> bool:
        """Check if node exist"""
        await self.run("SELECT ip FROM yascheduler_nodes WHERE ip=:ip;", ip=ip_addr)
        return bool(self.conn.row_count)

    async def update_task_status(self, task_id: int, status: TaskStatus) -> None:
        """Update task status"""
        await self.run(
            "UPDATE yascheduler_tasks SET status=:status WHERE task_id=:task_id;",
            task_id=task_id,
            status=status.value,
        )

    async def get_all_nodes(self) -> Sequence[NodeModel]:
        """Get all nodes"""
        rows = await self.run(
            """SELECT ip, ncpus, enabled, cloud, username FROM yascheduler_nodes;"""
        )
        return [NodeModel(*x) for x in (rows or [])]

    async def get_enabled_nodes(self) -> Sequence[NodeModel]:
        """Get all enabled nodes"""
        rows = await self.run(
            """SELECT ip, ncpus, enabled, cloud, username
            FROM yascheduler_nodes WHERE enabled=TRUE;"""
        )
        return [y for y in [NodeModel(*x) for x in (rows or [])] if "." in y.ip]

    async def get_disabled_nodes(self) -> Sequence[NodeModel]:
        """Get all disabled nodes"""
        rows = await self.run(
            """SELECT ip, ncpus, enabled, cloud, username
            FROM yascheduler_nodes WHERE enabled=FALSE;"""
        )
        return [y for y in [NodeModel(*x) for x in (rows or [])] if "." in y.ip]

    async def get_node(self, ip_addr: str) -> Optional[NodeModel]:
        """Get node by ip"""
        rows = await self.run(
            """SELECT ip, ncpus, enabled, cloud, username
            FROM yascheduler_nodes
            WHERE ip=:ip;""",
            ip=ip_addr,
        )
        for row in rows or []:
            return NodeModel(*row)

    async def count_nodes_clouds(self) -> Mapping[str, int]:
        """Count nodes in clouds"""
        rows = await self.run(
            """SELECT cloud, COUNT(cloud) FROM yascheduler_nodes
            WHERE cloud IS NOT NULL GROUP BY cloud;"""
        )
        data = {}
        for row in rows or []:
            data[row[0]] = row[1]
        return data

    async def count_nodes_by_status(self) -> Mapping[bool, int]:
        """Count nodes by status"""
        rows = await self.run(
            """SELECT enabled, COUNT(ip) FROM yascheduler_nodes
            GROUP BY enabled ORDER BY enabled;"""
        )
        data = defaultdict(lambda: 0)
        for row in rows or []:
            data[bool(row[0])] = row[1]
        return data

    async def add_tmp_node(self, cloud: str, username: str) -> str:
        """Add temporary node"""
        rows = await self.run(
            """INSERT INTO yascheduler_nodes (ip, enabled, cloud, username)
            VALUES ('prov' || SUBSTR(MD5(RANDOM()::TEXT), 0, 11),
              FALSE, :cloud, :username)
            RETURNING ip;""",
            cloud=cloud,
            username=username,
        )
        rows = cast(List[List[str]], rows)
        return rows[0][0]

    async def add_node(
        self,
        ip_addr: str,
        username: str,
        ncpus: Optional[int] = None,
        cloud: Optional[str] = None,
        enabled: bool = False,
    ) -> NodeModel:
        """Add new node"""
        await self.run(
            """INSERT INTO yascheduler_nodes (ip, ncpus, enabled, cloud, username)
            VALUES (:ip, :ncpus, :enabled, :cloud, :username);""",
            ip=ip_addr,
            ncpus=ncpus,
            cloud=cloud,
            username=username,
            enabled=enabled,
        )
        return NodeModel(
            ip_addr, ncpus, enabled=enabled, cloud=cloud, username=username
        )

    async def enable_node(self, ip_addr: str) -> None:
        """Enable node"""
        await self.run(
            "UPDATE yascheduler_nodes SET enabled=TRUE WHERE ip=:ip;",
            ip=ip_addr,
        )

    async def disable_node(self, ip_addr: str) -> None:
        """Disable node"""
        await self.run(
            "UPDATE yascheduler_nodes SET enabled=FALSE WHERE ip=:ip;",
            ip=ip_addr,
        )

    async def remove_node(self, ip_addr: str) -> None:
        """Remove node"""
        await self.run("DELETE FROM yascheduler_nodes WHERE ip=:ip;", ip=ip_addr)

    async def get_task(self, task_id: int) -> Optional[TaskModel]:
        """Get task"""
        rows = await self.run(
            """SELECT task_id, label, ip, status, metadata
                FROM yascheduler_tasks
                WHERE task_id=:task_id;""",
            task_id=task_id,
        )
        for row in rows or []:
            return TaskModel(*row)

    async def get_task_ids_by_ip_and_status(
        self, ip_addr: str, status: TaskStatus
    ) -> Sequence[int]:
        """Get task ids by ip and status"""
        rows = await self.run(
            """SELECT task_id FROM yascheduler_tasks
            WHERE ip=:ip AND status=:status
            ORDER BY task_id;""",
            ip=ip_addr,
            status=status.value,
        )
        return [cast(int, x[0]) for x in (rows or [])]

    async def get_tasks_by_jobs(self, jobs: Sequence[int]) -> Sequence[TaskModel]:
        """Get tasks by ids"""
        rows = await self.run(
            """SELECT task_id, label, ip, status, metadata
            FROM yascheduler_tasks
            WHERE task_id IN (SELECT unnest(CAST (:task_ids AS int[]))) ORDER BY task_id;""",
            task_ids=jobs,
        )
        return [TaskModel(*x) for x in (rows or [])]

    async def get_tasks_by_status(
        self, statuses: Sequence[TaskStatus], limit: Optional[int] = None
    ) -> Sequence[TaskModel]:
        """Get tasks by status"""
        rows = await self.run(
            """SELECT task_id, label, ip, status, metadata
            FROM yascheduler_tasks
            WHERE status IN (SELECT unnest(CAST (:statuses AS int[]))) ORDER BY task_id
            LIMIT :lim;""",
            statuses=[x.value for x in statuses],
            lim=limit,
        )
        return [TaskModel(*x) for x in (rows or [])]

    async def get_tasks_with_cloud_by_id_status(
        self, ids: Sequence[int], status: TaskStatus
    ) -> Sequence[TaskModel]:
        """Get tasks with cloud by id and status"""
        rows = await self.run(
            """SELECT t.task_id, t.label, t.ip, t.status, t.metadata, n.cloud
            FROM yascheduler_tasks AS t
            JOIN yascheduler_nodes AS n ON n.ip=t.ip
            WHERE status=:status AND
            task_id IN (SELECT unnest(CAST (:ids AS int[]))) ORDER BY task_id;""",
            ids=ids,
            status=status.value,
        )
        return [TaskModel(*x) for x in (rows or [])]

    async def count_tasks_by_status(self) -> Mapping[TaskStatus, int]:
        """Count tasks by status"""
        rows = await self.run(
            """SELECT status, COUNT(task_id) FROM yascheduler_tasks
            GROUP BY status ORDER BY status;"""
        )
        data = defaultdict(lambda: 0)
        for row in rows or []:
            data[TaskStatus(row[0])] = row[1]
        return data

    async def add_task(
        self,
        label: Optional[str] = None,
        ip_addr: Optional[str] = None,
        status: TaskStatus = TaskStatus.TO_DO,
        metadata: Optional[Mapping[str, Any]] = None,
    ) -> TaskModel:
        """Add new task"""
        rows = await self.run(
            """INSERT INTO yascheduler_tasks (label, metadata, ip, status)
            VALUES (:label, :metadata, :ip, :status)
            RETURNING task_id, label, ip, status, metadata;""",
            label=label or "",
            metadata=metadata,
            ip=ip_addr,
            status=status.value,
        )
        return TaskModel(*cast(list, rows)[0])

    async def update_task_meta(self, task_id: int, metadata: Mapping[str, Any]):
        """Update task metadata"""
        await self.run(
            "UPDATE yascheduler_tasks SET metadata=:metadata WHERE task_id=:task_id;",
            task_id=task_id,
            metadata=metadata,
        )

    async def set_task_running(self, task_id: int, ip_addr: str):
        """Set task running"""
        await self.run(
            """UPDATE yascheduler_tasks
            SET status=:status, ip=:ip
            WHERE task_id=:task_id;""",
            task_id=task_id,
            status=TaskStatus.RUNNING.value,
            ip=ip_addr,
        )

    async def set_task_done(self, task_id: int, metadata: Mapping[str, Any]):
        """Set task done"""
        await self.run(
            """UPDATE yascheduler_tasks
            SET status=:status, metadata=:metadata
            WHERE task_id=:task_id;""",
            task_id=task_id,
            metadata=metadata,
            status=TaskStatus.DONE.value,
        )

    async def set_task_error(
        self, task_id: int, metadata: Mapping[str, Any], error: Optional[str] = None
    ):
        """Set task error"""
        new_meta = (
            dict(list(metadata.items()) + [("error", error)]) if error else metadata
        )
        await self.run(
            """UPDATE yascheduler_tasks
            SET status=:status, metadata=:metadata
            WHERE task_id=:task_id;""",
            task_id=task_id,
            metadata=new_meta,
            status=TaskStatus.DONE.value,
        )



#!/usr/bin/env python3

import logging
import re
from pathlib import PurePath
from typing import AsyncGenerator, Optional, Pattern, Sequence, Union

from asyncssh.connection import SSHClientConnection
from asyncssh.sftp import SFTPClient

from ..config import LocalArchiveDeploy, LocalFilesDeploy, RemoteArchiveDeploy
from .common import ProcessInfo
from .protocol import OuterRunCallable, PEngineRepository, PProcessInfo, QuoteCallable


async def linux_get_cpu_cores(run: OuterRunCallable) -> int:
    """
    Get number of CPU cores
    :raises asyncssh.Error: An SSH error has occurred.
    """
    r = await run("getconf NPROCESSORS_ONLN 2> /dev/null || getconf _NPROCESSORS_ONLN")
    try:
        return int(r.stdout and r.stdout.strip() or "1")
    except ValueError:
        return 1


async def linux_list_processes(
    conn: SSHClientConnection, query: Optional[str] = None
) -> AsyncGenerator[PProcessInfo, None]:
    """
    Returns information about all running processes
    :raises asyncssh.Error: An SSH error has occurred.
    """
    columns = ["pid", "comm", "args"]
    columns_part = ",".join([f"{x}:255" for x in columns])
    if query:
        ps_cmd = " ".join([query, "| xargs --no-run-if-empty ps -o", columns_part])
    else:
        ps_cmd = f"ps -eo {columns_part}"
    async with conn.create_process(ps_cmd) as proc:
        await proc.stdout.readline()  # skip headers
        async for line in proc.stdout:
            parts = list(
                map(lambda x: x.strip(), filter(None, str(line).split(" " * 10)))
            )
            # skip broken
            if len(parts) < 3:
                continue
            # skip parent of self
            if parts[2].startswith(f"bash -c {ps_cmd}"):
                continue
            yield ProcessInfo(int(parts[0]), *parts[1:3])


async def linux_pgrep(
    conn: SSHClientConnection,
    quote: QuoteCallable,
    pattern: Union[str, Pattern[str]],
    full=True,
) -> AsyncGenerator[PProcessInfo, None]:
    """
    Returns information about running processes, that name matches a pattern.
    If `full`, check match against name or full cmd.
    :raises asyncssh.Error: An SSH error has occurred.
    """
    str_pattern = pattern.pattern if isinstance(pattern, re.Pattern) else pattern
    pgrep_query = " ".join(
        filter(None, ["pgrep", "-f" if full else None, quote(str_pattern)])
    )
    async for x in linux_list_processes(conn, query=pgrep_query):
        yield x


async def deploy_local_files(
    sftp: SFTPClient,
    engine_dir: PurePath,
    files: Sequence[PurePath],
    log: Optional[logging.Logger] = None,
):
    "Uploading binary from local; requires broadband connection"
    lpaths = list(map(str, files))
    if log:
        log.debug(f"Uploading files ({', '.join(lpaths)}) to {engine_dir}")
    await sftp.put(lpaths, engine_dir, preserve=True)


async def deploy_local_archive(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engine_dir: PurePath,
    archive: PurePath,
    log: Optional[logging.Logger] = None,
):
    """
    Upload local archive.
    Binary may be gzipped, without subfolders, with an arbitrary archive name.
    """
    rpath = engine_dir / archive.name
    if log:
        log.debug(f"Uploading {archive.name} to {str(rpath)}...")
    await sftp.put([str(archive)], engine_dir)
    if log:
        log.debug(f"Unarchiving {archive.name}...")
    await run(f"tar xfv {quote(str(archive.name))}", cwd=str(engine_dir), check=True)
    await sftp.remove(rpath)


async def deploy_remote_archive(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engine_dir: PurePath,
    url: str,
    log: Optional[logging.Logger] = None,
):
    """
    Downloading binary from a trusted non-public address.
    Binary may be gzipped, without subfolders, with an arbitrary archive name.
    """
    name = "archive.tar.gz"
    rpath = engine_dir / name
    if log:
        log.debug(f"Downloading {url} to {str(rpath)}...")
    await run(f"wget {quote(url)} -O {quote(name)}", cwd=str(engine_dir), check=True)
    if log:
        log.debug(f"Unarchiving {name}...")
    await run(f"tar xfv {quote(str(name))}", cwd=str(engine_dir), check=True)
    await sftp.remove(rpath)


async def linux_deploy_engines(
    run: OuterRunCallable,
    quote: QuoteCallable,
    sftp: SFTPClient,
    engines: PEngineRepository,
    engines_dir: PurePath,
    log: Optional[logging.Logger] = None,
) -> None:
    """
    Setup node for target engines.
    """
    for engine in engines.values():
        if log:
            log.info(f"Setup {engine.name} engine...")
        engine_dir = engines_dir / engine.name
        await sftp.makedirs(engine_dir, exist_ok=True)
        for deployment in engine.deployable:
            if isinstance(deployment, LocalFilesDeploy):
                await deploy_local_files(sftp, engine_dir, deployment.files, log)

            if isinstance(deployment, LocalArchiveDeploy):
                await deploy_local_archive(
                    run, quote, sftp, engine_dir, deployment.file
                )

            if isinstance(deployment, RemoteArchiveDeploy):
                await deploy_remote_archive(
                    run, quote, sftp, engine_dir, deployment.url
                )
        if log:
            log.info(f"Setup of {engine.name} engine is done...")


async def log_mpi_version(run: OuterRunCallable, log: Optional[logging.Logger] = None):
    r = await run("mpirun --allow-run-as-root -V", check=True)
    if not r.returncode and log:
        log.debug(str(r.stdout or "").split("\n")[0])


async def linux_setup_node(
    conn: SSHClientConnection,
    run: OuterRunCallable,
    quote: QuoteCallable,
    engines: PEngineRepository,
    engines_dir: PurePath,
    log: Optional[logging.Logger] = None,
):
    "Setup generic linux node"
    async with conn.start_sftp_client() as sftp:
        await linux_deploy_engines(run, quote, sftp, engines, engines_dir, log)


async def linux_setup_deb_node(
    conn: SSHClientConnection,
    run: OuterRunCallable,
    quote: QuoteCallable,
    engines: PEngineRepository,
    engines_dir: PurePath,
    log: Optional[logging.Logger] = None,
):
    "Setup debian-like node"
    is_root = conn._username == "root"
    sudo_prefix = "" if is_root else "sudo "
    apt_cmd = f"{sudo_prefix}apt-get -o DPkg::Lock::Timeout=600 -y"
    pkgs = engines.get_platform_packages()

    if log:
        log.debug("Upgrade packages...")
    await run(f"{apt_cmd} update", check=True)
    await run(f"{apt_cmd} upgrade", check=True)
    if pkgs:
        if log:
            log.debug("Install packages: {} ...".format(" ".join(pkgs)))
        await run(f"{apt_cmd} install {' '.join(pkgs)}", check=True)
    if [x for x in pkgs if "mpi" in x]:
        await log_mpi_version(run, log)

    async with conn.start_sftp_client() as sftp:
        await linux_deploy_engines(run, quote, sftp, engines, engines_dir, log)



"Remote machine"

import asyncio
import logging
from asyncio.locks import Event, Semaphore
from contextlib import asynccontextmanager
from datetime import datetime, timedelta
from functools import partial
from pathlib import PurePath
from typing import AsyncGenerator, Optional, Pattern, Sequence, Set, Type, Union

import asyncssh
import backoff
from asyncssh.client import SSHClient
from asyncssh.connection import SSHClientConnection, SSHClientConnectionOptions
from asyncssh.process import SSHClientProcess, SSHCompletedProcess
from asyncssh.public_key import SSHKey
from asyncssh.sftp import SFTPClient
from asyncstdlib import all as aall
from asyncstdlib import map as amap
from attrs import define, field, validators
from typing_extensions import Self

from .adapters import (
    darwin_adapter,
    debian_10_adapter,
    debian_11_adapter,
    debian_12_adapter,
    debian_13_adapter,
    debian_14_adapter,
    debian_15_adapter,
    debian_adapter,
    debian_like_adapter,
    linux_adapter,
    windows7_adapter,
    windows8_adapter,
    windows10_adapter,
    windows11_adapter,
    windows12_adapter,
    windows_adapter,
)
from .exc import PlatformGuessFailed
from .protocol import (
    AllSSHRetryExc,
    PEngine,
    PEngineRepository,
    PProcessInfo,
    PRemoteMachine,
    PRemoteMachineAdapter,
    PRemoteMachineMetadata,
    SSHCheck,
    SSHRetryExc,
)

ADAPTERS: Sequence[PRemoteMachineAdapter] = [
    debian_10_adapter,
    debian_11_adapter,
    debian_12_adapter,
    debian_13_adapter,
    debian_14_adapter,
    debian_15_adapter,
    debian_adapter,
    debian_like_adapter,
    linux_adapter,
    darwin_adapter,
    windows10_adapter,
    windows11_adapter,
    windows12_adapter,
    windows7_adapter,
    windows8_adapter,
    windows_adapter,
]

# default value of MaxSessions on OpenSSH server is 10
MAX_SESSIONS = 10

my_backoff_exc = partial(
    backoff.on_exception,
    wait_gen=backoff.fibo,
    max_time=60,
    exception=SSHRetryExc,
)


class MySSHClient(SSHClient):
    def validate_host_public_key(
        self, host: str, addr: str, port: int, key: SSHKey
    ) -> bool:
        "Trust all host keys"
        # NOTE: this is insecure for MiM attacks
        return True


DEFAULT_CONN_OPTS = SSHClientConnectionOptions(
    client_factory=MySSHClient,
    preferred_auth="publickey",
    keepalive_interval=10,
    keepalive_count_max=10,
    compression_algs=[],
    agent_path="",
    username="root",
)


@define
class RemoteMachineMetadata(PRemoteMachineMetadata):
    def __init__(self):
        self._busy = None
        self._free_since: Optional[datetime] = datetime.now()

    @property
    def busy(self) -> Optional[bool]:
        return self._busy

    @busy.setter
    def busy(self, new_busy: bool):
        if new_busy:
            self._busy = True
            self._free_since = None
        else:
            self._busy = False
            self._free_since = datetime.now()

    def is_free_longer_than(self, delta: timedelta) -> bool:
        if not self._free_since or self.busy:
            return False
        return datetime.now() - delta > self._free_since


@define(frozen=True)
class RemoteMachine(PRemoteMachine):
    "Remote SSH machine"

    conn: SSHClientConnection = field()
    conn_opts: SSHClientConnectionOptions = field()

    meta: RemoteMachineMetadata = field()

    adapter: PRemoteMachineAdapter = field()
    log: logging.Logger = field()

    platforms: Sequence[str] = field()

    data_dir: PurePath = field(validator=validators.instance_of(PurePath))
    engines_dir: PurePath = field(validator=validators.instance_of(PurePath))
    tasks_dir: PurePath = field(validator=validators.instance_of(PurePath))

    cancellation_event: Event = field(factory=Event, init=False)
    jobs: Set[asyncio.Task] = field(factory=set, init=False)
    sessions_limit: Semaphore = field(
        factory=lambda: Semaphore(MAX_SESSIONS), init=False
    )

    def __le__(self, other: Self) -> bool:
        if not self.meta._free_since:
            return True
        if not other.meta._free_since:
            return False
        return self.meta._free_since <= other.meta._free_since

    def __gt__(self, other: Self) -> bool:
        if not self.meta._free_since:
            return False
        if not other.meta._free_since:
            return True
        return self.meta._free_since > other.meta._free_since

    @classmethod
    @my_backoff_exc()
    async def create(
        cls,
        host: str,
        username: str,
        client_keys: Optional[Sequence[PurePath]],
        logger: Optional[logging.Logger] = None,
        connect_timeout: Optional[int] = None,
        data_dir: Optional[PurePath] = None,
        engines_dir: Optional[PurePath] = None,
        tasks_dir: Optional[PurePath] = None,
        jump_host: Optional[str] = None,
        jump_username: Optional[str] = None,
    ) -> "PRemoteMachine":
        logger_name = f"{cls.__name__}:{username}@{host}"
        if logger:
            log = logger.getChild(logger_name)
        else:
            log = logging.getLogger(logger_name)
        # If our logging level is not set then asyncssh is too noisy
        asyncssh.logging.set_log_level(logging.WARNING if log.level == 0 else log.level)
        # asyncssh.logging.set_log_level("DEBUG")
        # asyncssh.logging.set_debug_level(2)

        # connection
        conn_opts = SSHClientConnectionOptions(
            options=DEFAULT_CONN_OPTS,
            host=host,
            username=username,
            tunnel=jump_host and jump_username and f"{jump_username}@{jump_host}",
            client_keys=client_keys or (),
            ignore_encrypted=True,
            connect_timeout=connect_timeout,
        )

        log.debug("Open connection")
        conn = await asyncssh.connection.connect(
            options=conn_opts,
            host=conn_opts.host,
            tunnel=conn_opts.tunnel,
        )

        # guess platform
        sess_lim = Semaphore(MAX_SESSIONS)

        async def with_limit(conn: SSHClientConnection, fn: SSHCheck):
            async with sess_lim:
                return await fn(conn)

        adapter = None
        platforms: Sequence[str] = []
        checks: Sequence[bool] = [
            await aall(amap(lambda y: with_limit(conn, y), x.checks)) for x in ADAPTERS
        ]

        for candidate, check in zip(ADAPTERS, checks):  # noqa: B905
            if check:
                platforms.append(candidate.platform)
            if check and not adapter:
                adapter = candidate

        if not adapter:
            raise PlatformGuessFailed()

        log.debug(f"Detected platform: {adapter.platform}")

        Path = adapter.path
        if not isinstance(data_dir, Path):
            data_dir = Path(str(data_dir)) if data_dir else Path("./data")
        if not isinstance(engines_dir, Path):
            engines_dir = (
                Path(str(engines_dir)) if engines_dir else data_dir / "engines"
            )
        if not isinstance(tasks_dir, Path):
            tasks_dir = Path(str(tasks_dir)) if tasks_dir else data_dir / "tasks"

        return cls(
            conn=conn,
            conn_opts=conn_opts,
            meta=RemoteMachineMetadata(),
            adapter=adapter,
            platforms=platforms,
            log=log,
            data_dir=data_dir,
            engines_dir=engines_dir,
            tasks_dir=tasks_dir,
        )

    @classmethod
    @asynccontextmanager
    async def create_ctx(
        cls, *args, **kwargs
    ) -> AsyncGenerator["PRemoteMachine", None]:
        """
        Create async context.
        :raises asyncssh.Error: An SSH error has occurred.
        """
        machine = await cls.create(*args, **kwargs)
        yield machine
        await machine.close()

    @property
    def hostname(self) -> str:
        return self.conn_opts.host

    async def close(self) -> None:
        "Close connections and free resources"
        self.cancellation_event.set()
        for task in self.jobs:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
        if self.conn._transport:
            self.log.debug("Close connection")
            self.conn.close()
            await self.conn.wait_closed()

    @asynccontextmanager
    async def sftp(self, **kwargs) -> AsyncGenerator[SFTPClient, None]:
        """
        Open SFTP connection
        :raises asyncssh.SFTPError: An SFTP error has occurred.
        """
        conn = await self.get_conn()
        async with conn.start_sftp_client(**kwargs) as sftp:
            yield sftp

    @my_backoff_exc()
    async def renew_conn(self) -> SSHClientConnection:
        """
        Reopen SSH connection
        """
        conn = await asyncssh.connection.connect(
            options=self.conn_opts,
            host=self.conn_opts.host,
            tunnel=self.conn_opts.tunnel,
        )
        # MUTATE OBJECT!
        object.__setattr__(self, "conn", conn)
        return conn

    async def get_conn(self) -> SSHClientConnection:
        """
        Return current SSH connection. Reopen if needed.
        """
        async with self.sessions_limit:
            if self.conn._transport and not self.conn._transport.is_closing():
                return self.conn
            self.log.debug("Connection is closed - reopening")
            return await self.renew_conn()

    @property
    def path(self) -> Type[PurePath]:
        "Return path of the adapter"
        return self.adapter.path

    def quote(self, string: str) -> str:
        "Platform-specific shell quoting"
        return self.adapter.quote(string)

    @my_backoff_exc()
    async def run(
        self, *args, cwd: Optional[str] = None, **kwargs
    ) -> SSHCompletedProcess:
        "Run process and wait for exit"
        conn = await self.get_conn()
        return await self.adapter.run(
            conn, self.adapter.quote, *args, cwd=cwd, **kwargs
        )

    async def run_bg(
        self, command: str, *args, cwd: Optional[str] = None, **kwargs
    ) -> SSHClientProcess:
        "Run process in background"
        conn = await self.get_conn()
        return await self.adapter.run_bg(
            conn, self.adapter.quote, command, *args, cwd=cwd, **kwargs
        )

    @my_backoff_exc()
    async def get_cpu_cores(self) -> int:
        "Get number of CPU cores"
        return await self.adapter.get_cpu_cores(self.run)

    async def list_processes(self) -> AsyncGenerator[PProcessInfo, None]:
        "Returns information about all running processes"
        conn = await self.get_conn()
        async for x in self.adapter.list_processes(conn, None):
            yield x

    async def pgrep(
        self, pattern: Union[str, Pattern], full: bool = True
    ) -> AsyncGenerator[PProcessInfo, None]:
        """
        Returns information about running processes, that name matches a pattern.
        If `full`, check match against name or full cmd.
        """
        conn = await self.get_conn()
        async for x in self.adapter.pgrep(conn, self.adapter.quote, pattern, full):
            yield x

    async def setup_node(self, engines: PEngineRepository):
        """
        Setup node for target engines.
        :raises NotImplemented: Not supported on platform.
        """
        self.log.info(f"CPUs count: {await self.get_cpu_cores()}")
        conn = await self.get_conn()
        # pylint: disable=redundant-keyword-arg
        retry = my_backoff_exc(exception=AllSSHRetryExc)
        await retry(self.adapter.setup_node)(
            conn=conn,
            run=self.run,
            quote=self.quote,
            engines=engines.filter_platforms(self.platforms),
            engines_dir=self.engines_dir,
            log=self.log,
        )

    async def occupancy_check(self, engine: PEngine) -> bool:
        """
        Check if a node is occupied by an engine's task
        """
        if engine.check_pname:
            try:
                if [x async for x in self.pgrep(engine.check_pname)]:
                    return True
            except SSHRetryExc as exc:
                self.log.info(f"Node {self.hostname} failed pgrep: {exc}")
                await self.renew_conn()
        elif engine.check_cmd:
            try:
                proc = await self.run(engine.check_cmd)
                if proc.returncode == engine.check_cmd_code:
                    return True
            except SSHRetryExc as exc:
                self.log.info(f"Node {self.hostname} failed command: {exc}")
                await self.renew_conn()
        return False

    async def start_occupancy_check(self, engine: PEngine) -> None:
        """
        Start occupancy checker for engine.
        """
        # remove old tasks
        self.jobs.difference_update([t for t in self.jobs if t.done()])

        async def occupancy_checker():
            while not self.cancellation_event.is_set() and self.meta.busy is not False:
                await asyncio.sleep(engine.sleep_interval)

                try:
                    self.meta.busy = await asyncio.wait_for(
                        self.occupancy_check(engine), timeout=engine.sleep_interval
                    )
                except asyncio.TimeoutError:
                    tmpl = "Engine {} busy check timeouted on {}"
                    self.log.warning(tmpl.format(engine.name, self.hostname))
                except Exception as err:  # pylint: disable=broad-exception-caught
                    self.log.warning(err)

        self.jobs.add(asyncio.create_task(occupancy_checker()))



"""Cloud API module"""

import asyncio
import base64
import json
import logging
from typing import Optional, Sequence, Union

import backoff
from asyncssh.process import ProcessError
from asyncssh.public_key import SSHKey, generate_private_key, read_private_key
from attrs import asdict, define, field

from ..config import ConfigLocal, ConfigRemote, EngineRepository
from ..remote_machine import PRemoteMachine, RemoteMachine, SSHRetryExc
from .protocols import PCloudAdapter, PCloudAPI, PCloudConfig, TConfigCloud_contra
from .utils import get_rnd_name


class CloudCreateNodeError(Exception):
    """Cloud node allocation error"""


class CloudSetupNodeError(Exception):
    """Cloud node setup error"""


@define(frozen=True)
class CloudConfig(PCloudConfig):
    "Cloud config init"

    bootcmd: Sequence[Union[str, Sequence[str]]] = field(factory=tuple)
    package_upgrade: bool = field(default=False)
    packages: Sequence[str] = field(factory=list)

    def render(self) -> str:
        "Render to user-data format"
        return "#cloud-config\n" + json.dumps(asdict(self))

    def render_base64(self) -> str:
        "Render to user-data format as base64 string"
        return base64.b64encode(self.render().encode()).decode()


@define(frozen=True)
class CloudAPI(PCloudAPI[TConfigCloud_contra]):
    "Cloud API protocol"

    adapter: PCloudAdapter[TConfigCloud_contra] = field()
    config: TConfigCloud_contra = field()
    local_config: ConfigLocal = field()
    remote_config: ConfigRemote = field()
    engines: EngineRepository = field()
    log: logging.Logger = field()
    ssh_key_lock: asyncio.Lock = field(factory=asyncio.Lock)

    @property
    def name(self) -> str:
        "Cloud name"
        return self.adapter.name

    @classmethod
    async def create(
        cls,
        adapter: PCloudAdapter,
        config: TConfigCloud_contra,
        local_config: ConfigLocal,
        remote_config: ConfigRemote,
        engines: EngineRepository,
        log: Optional[logging.Logger] = None,
        ssh_key_lock: Optional[asyncio.Lock] = None,
    ):
        "Create cloud API"
        if log:
            log = log.getChild(adapter.name)
        else:
            log = logging.getLogger(adapter.name)

        return cls(
            adapter=adapter,
            config=config,
            local_config=local_config,
            remote_config=remote_config,
            engines=engines,
            log=log,
            ssh_key_lock=ssh_key_lock or asyncio.Lock(),
        )

    def get_op_semaphore(self) -> asyncio.Semaphore:
        return self.adapter.get_op_semaphore()

    def is_platform_supported(self, platform: str) -> bool:
        return any(map(lambda x: x(platform), self.adapter.supported_platform_checks))

    def get_ssh_key_sync(self) -> SSHKey:
        "Load or generate new SSHKey"
        prefix = "yakey"
        # try to load
        for filepath in self.local_config.keys_dir.iterdir():
            if not filepath.name.startswith(prefix) or not filepath.is_file():
                continue
            ssh_key = read_private_key(filepath)
            ssh_key.set_comment(filepath.name)
            self.log.debug(
                "LOADED KEY %s: %s", filepath.name, ssh_key.get_fingerprint("md5")
            )
            return ssh_key

        key_name = get_rnd_name(prefix)
        filepath = self.local_config.keys_dir / key_name
        ssh_key = generate_private_key(alg_name="ssh-rsa", comment=key_name)
        ssh_key.write_private_key(filepath)
        filepath.chmod(0o600)
        ssh_key.set_comment(key_name)
        self.log.info("WRITTEN KEY %s: %s", key_name, ssh_key.get_fingerprint("md5"))
        return ssh_key

    async def get_ssh_key(self) -> SSHKey:
        "Load or generate ssh key (cached)"
        async with self.ssh_key_lock:
            return await asyncio.get_running_loop().run_in_executor(
                None, self.get_ssh_key_sync
            )

    async def get_cloud_config_data(self) -> PCloudConfig:
        "Common cloud-config"
        engines = self.engines.filter(
            lambda e: bool(e.platforms)
            and any(map(self.is_platform_supported, e.platforms))
            or not e.platforms
        )
        pkgs = engines.get_platform_packages()
        return CloudConfig(package_upgrade=True, packages=pkgs)

    async def mk_machine(self, ip_addr: str) -> PRemoteMachine:
        "Create RemoteMachine"
        keys = await asyncio.get_running_loop().run_in_executor(
            None, self.local_config.get_private_keys
        )
        retry = backoff.on_exception(
            wait_gen=backoff.fibo,
            max_time=self.adapter.create_node_timeout,
            exception=SSHRetryExc,
        )
        return await retry(RemoteMachine.create)(
            host=ip_addr,
            username=self.config.username,
            client_keys=keys,
            logger=self.log,
            connect_timeout=self.adapter.create_node_conn_timeout,
            data_dir=self.remote_config.data_dir,
            engines_dir=self.remote_config.engines_dir,
            tasks_dir=self.remote_config.tasks_dir,
            jump_host=self.config.jump_host,
            jump_username=self.config.jump_username,
        )

    async def create_node(self):
        async with self.adapter.get_op_semaphore():
            try:
                ip_addr = await self.adapter.create_node(
                    log=self.log,
                    cfg=self.config,
                    key=await self.get_ssh_key(),
                    cloud_config=await self.get_cloud_config_data(),
                )
            except Exception as err:
                raise CloudCreateNodeError(f"Create node error: {err}") from err

            try:
                machine = await self.mk_machine(ip_addr)
                await machine.run("cloud-init status --wait")
                await machine.setup_node(self.engines)
            except (ProcessError, Exception) as err:
                if isinstance(err, ProcessError):
                    self.log.error(
                        (
                            "Setup node %s failed: command '%s' failed "
                            "with exit code %s; stderr: %s"
                        ),
                        ip_addr,
                        err.command,
                        err.returncode,
                        err.stderr,
                    )
                self.log.warn("Setup node %s failed - deallocate", ip_addr)
                await self.delete_node(ip_addr)
                raise CloudSetupNodeError(f"Setup node error: {err}") from err
            return ip_addr

    async def delete_node(self, host: str):
        async with self.adapter.get_op_semaphore():
            return await self.adapter.delete_node(
                log=self.log, cfg=self.config, host=host
            )



#!/usr/bin/env python3
from .protocol import (
    AllSSHRetryExc,
    PProcessInfo,
    PRemoteMachine,
    SFTPRetryExc,
    SSHRetryExc,
)
from .remote_machine import RemoteMachine
from .remote_machine_repository import RemoteMachineRepository

__all__ = [
    "AllSSHRetryExc",
    "PProcessInfo",
    "PRemoteMachine",
    "RemoteMachine",
    "RemoteMachineRepository",
    "SFTPRetryExc",
    "SSHRetryExc",
]



import importlib.metadata

from .client import Yascheduler
from .variables import CONFIG_FILE, LOG_FILE, PID_FILE

__version__ = importlib.metadata.version("yascheduler")
__all__ = [
    "CONFIG_FILE",
    "LOG_FILE",
    "PID_FILE",
    "Yascheduler",
    "__version__",
]



#!/usr/bin/env python3
"""Engine configuration"""

from configparser import SectionProxy
from pathlib import PurePath
from typing import Optional, Sequence, Tuple, Union

from attrs import Attribute, define, field, fields, validators

from .utils import _make_default_field, warn_unknown_fields


def _check_spawn(instance: "Engine", _, value: str):
    try:
        value.format(task_path="", engine_path="", ncpus="")
    except KeyError as err:
        msg = (
            "Engine {name} has unknown template placeholder "
            "`{placeholder}` in *spawn* command"
        )
        raise ValueError(
            msg.format(name=instance.name, placeholder=err.args[0])
        ) from err


def _check_check_(instance: "Engine", attribute: Attribute, value: Optional[str]):
    no_check_cmd_curr = attribute.name == "check_cmd" and not value
    no_check_pname_curr = attribute.name == "check_pname" and not value
    if (no_check_cmd_curr and not instance.check_pname) or (
        no_check_pname_curr and not instance.check_cmd
    ):
        raise ValueError(
            f"Engine {instance.name} has no *check_cmd* or *check_pname* set"
        )


def _check_at_least_one_elem(
    instance: "Engine", attribute: Attribute, value: Optional[Sequence]
):
    if not value or len(value) < 1:
        raise ValueError(f"Engine {instance.name} has no *{attribute.name}* config set")


@define(frozen=True)
class LocalFilesDeploy:
    "Deploy local files configuration"

    files: Tuple[PurePath] = field(factory=tuple)


@define(frozen=True)
class LocalArchiveDeploy:
    "Deploy local archive configuration"

    file: PurePath


@define(frozen=True)
class RemoteArchiveDeploy:
    "Deploy remote archive configuration"

    url: str


Deploy = Union[
    LocalFilesDeploy,
    LocalArchiveDeploy,
    RemoteArchiveDeploy,
]


@define(frozen=True)
class Engine:
    """Engine configuration"""

    name: str = field(validator=[validators.instance_of(str)])
    spawn: str = field(validator=[validators.instance_of(str), _check_spawn])
    check_cmd: Optional[str] = field(
        validator=[validators.optional(validators.instance_of(str)), _check_check_]
    )
    check_pname: Optional[str] = field(
        validator=[validators.optional(validators.instance_of(str)), _check_check_]
    )
    deployable: Tuple[Deploy, ...] = field(factory=tuple)
    input_files: Tuple[str, ...] = field(
        factory=tuple,
        validator=[
            validators.deep_iterable(member_validator=validators.instance_of(str)),
            _check_at_least_one_elem,
        ],
    )
    output_files: Tuple[str, ...] = field(
        factory=tuple,
        validator=[
            validators.deep_iterable(member_validator=validators.instance_of(str)),
            _check_at_least_one_elem,
        ],
    )
    platforms: Tuple[str, ...] = field(
        factory=tuple,
        validator=[
            validators.deep_iterable(member_validator=validators.instance_of(str))
        ],
    )
    platform_packages: Tuple[str, ...] = field(
        factory=tuple,
        validator=[
            validators.deep_iterable(member_validator=validators.instance_of(str))
        ],
    )
    check_cmd_code: int = _make_default_field(0)
    sleep_interval: int = _make_default_field(10)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        exclude_names = ["name", "deployable"]
        include_names = [
            "deploy_local_files",
            "deploy_local_archive",
            "deploy_remote_archive",
        ]
        return [
            f.name for f in fields(cls) if f.name not in exclude_names
        ] + include_names

    @classmethod
    def from_config_parser_section(
        cls, sec: SectionProxy, engines_dir: PurePath
    ) -> "Engine":
        "Create config from config parser's section"

        warn_unknown_fields(cls.get_valid_config_parser_fields(), sec)

        def gettuple(key: str) -> Tuple[str]:
            return tuple(
                x.strip() for x in filter(None, sec.get(key, fallback="").split())
            )

        name = sec.name[7:]
        engine_dir = engines_dir / name

        deployable: Sequence[Deploy] = []
        deploy_local_files = [
            engine_dir / x.strip() for x in gettuple("deploy_local_files")
        ]
        if deploy_local_files:
            deployable.append(LocalFilesDeploy(files=tuple(deploy_local_files)))
        deploy_local_archive = sec.get("deploy_local_archive", None)
        if deploy_local_archive:
            deploy = LocalArchiveDeploy(file=engine_dir / deploy_local_archive)
            deployable.append(deploy)
        deploy_remote_archive = sec.get("deploy_remote_archive", None)
        if deploy_remote_archive:
            deploy = RemoteArchiveDeploy(url=deploy_remote_archive)
            deployable.append(deploy)

        return cls(
            name=name,
            deployable=tuple(deployable),
            spawn=sec.get("spawn"),
            check_cmd=sec.get("check_cmd"),
            check_cmd_code=sec.getint("check_cmd_code"),
            check_pname=sec.get("check_pname"),
            input_files=gettuple("input_files"),
            output_files=gettuple("output_files"),
            sleep_interval=sec.getint("sleep_interval"),
            platforms=gettuple("platforms"),
            platform_packages=gettuple("platform_packages"),
        )



#!/usr/bin/env python3

import shlex
from pathlib import PurePath, PurePosixPath
from typing import Sequence, Type

from attrs import define, evolve, field

from .checks import (
    check_is_darwin,
    check_is_debian,
    check_is_debian_10,
    check_is_debian_11,
    check_is_debian_12,
    check_is_debian_13,
    check_is_debian_14,
    check_is_debian_15,
    check_is_debian_like,
    check_is_linux,
    check_is_windows,
    check_is_windows7,
    check_is_windows8,
    check_is_windows10,
    check_is_windows11,
    check_is_windows12,
)
from .common import run, run_bg
from .linux_methods import (
    linux_get_cpu_cores,
    linux_list_processes,
    linux_pgrep,
    linux_setup_deb_node,
    linux_setup_node,
)
from .protocol import (
    GetCPUCoresCallable,
    ListProcessesCallable,
    PgrepCallable,
    PRemoteMachineAdapter,
    QuoteCallable,
    RunBgCallable,
    RunCallable,
    SetupNodeCallable,
    SSHCheck,
)
from .windows_methods import (
    MyPureWindowsPath,
    windows_get_cpu_cores,
    windows_list_processes,
    windows_pgrep,
    windows_quote,
    windows_setup_node,
)


@define(frozen=True)
class RemoteMachineAdapter(PRemoteMachineAdapter):
    "Remote machine adapter"

    platform: str = field()
    path: Type[PurePath] = field()

    quote: QuoteCallable = field()
    run: RunCallable = field()
    run_bg: RunBgCallable = field()
    get_cpu_cores: GetCPUCoresCallable = field()
    list_processes: ListProcessesCallable = field()
    pgrep: PgrepCallable = field()
    setup_node: SetupNodeCallable = field()

    checks: Sequence[SSHCheck] = field(factory=tuple)


linux_adapter = RemoteMachineAdapter(
    platform="linux",
    path=PurePosixPath,
    quote=shlex.quote,
    run=run,
    run_bg=run_bg,
    get_cpu_cores=linux_get_cpu_cores,
    list_processes=linux_list_processes,
    pgrep=linux_pgrep,
    setup_node=linux_setup_node,
    checks=(check_is_linux,),
)

debian_like_adapter = evolve(
    linux_adapter,
    platform="debian-like",
    setup_node=linux_setup_deb_node,
    checks=(*linux_adapter.checks, check_is_debian_like),
)

debian_adapter = evolve(
    debian_like_adapter,
    platform="debian",
    checks=(*debian_like_adapter.checks, check_is_debian),
)

debian_10_adapter = evolve(
    debian_adapter,
    platform="debian-10",
    checks=(*debian_adapter.checks, check_is_debian_10),
)

debian_11_adapter = evolve(
    debian_adapter,
    platform="debian-11",
    checks=(*debian_adapter.checks, check_is_debian_11),
)

debian_12_adapter = evolve(
    debian_adapter,
    platform="debian-12",
    checks=(*debian_adapter.checks, check_is_debian_12),
)

debian_13_adapter = evolve(
    debian_adapter,
    platform="debian-13",
    checks=(*debian_adapter.checks, check_is_debian_13),
)

debian_14_adapter = evolve(
    debian_adapter,
    platform="debian-14",
    checks=(*debian_adapter.checks, check_is_debian_14),
)

debian_15_adapter = evolve(
    debian_adapter,
    platform="debian-15",
    checks=(*debian_adapter.checks, check_is_debian_15),
)

darwin_adapter = evolve(
    linux_adapter,
    platform="darwin",
    checks=(check_is_darwin,),
)

windows_adapter = RemoteMachineAdapter(
    platform="windows",
    path=MyPureWindowsPath,
    quote=windows_quote,
    run=run,
    run_bg=run_bg,
    get_cpu_cores=windows_get_cpu_cores,
    list_processes=windows_list_processes,
    pgrep=windows_pgrep,
    setup_node=windows_setup_node,
    checks=(check_is_windows,),
)

windows7_adapter = evolve(
    windows_adapter,
    platform="windows-8",
    checks=(*windows_adapter.checks, check_is_windows7),
)

windows8_adapter = evolve(
    windows_adapter,
    platform="windows-8",
    checks=(*windows_adapter.checks, check_is_windows8),
)

windows10_adapter = evolve(
    windows_adapter,
    platform="windows-10",
    checks=(*windows_adapter.checks, check_is_windows10),
)

windows11_adapter = evolve(
    windows_adapter,
    platform="windows-11",
    checks=(*windows_adapter.checks, check_is_windows11),
)

windows12_adapter = evolve(
    windows_adapter,
    platform="windows-12",
    checks=(*windows_adapter.checks, check_is_windows12),
)



"OS checks"

from functools import partial
from typing import Optional, Tuple

from asyncssh.connection import SSHClientConnection
from asyncstdlib import lru_cache


@lru_cache
async def check_is_linux(conn: SSHClientConnection) -> bool:
    "Check for generic Linux"
    proc = await conn.run("uname")
    return (
        proc.returncode == 0
        and proc.stdout is not None
        and proc.stdout.strip() == "Linux"
    )


@lru_cache
async def check_is_darwin(conn: SSHClientConnection) -> bool:
    "Check for Mac"
    proc = await conn.run("uname")
    return (
        proc.returncode == 0
        and proc.stdout is not None
        and proc.stdout.strip() == "Darwin"
    )


@lru_cache
async def _get_os_release(conn: SSHClientConnection) -> Optional[Tuple[str]]:
    "Get os release string on linuxes"
    proc = await conn.run(
        "sh -c 'source /etc/os-release; echo $ID@@@$ID_LIKE@@@$VERSION_ID'"
    )
    if proc.returncode != 0 or not proc.stdout:
        return None
    return tuple(map(lambda x: x.strip(), str(proc.stdout).split("@@@", maxsplit=3)))


async def check_is_debian_like(conn: SSHClientConnection) -> bool:
    "Check for any Debian-like"
    os_release = await _get_os_release(conn)
    return "debian" in os_release[0:2] if os_release else False


async def check_is_debian(conn: SSHClientConnection) -> bool:
    "Check for any Debian"
    os_release = await _get_os_release(conn)
    return os_release[0] == "debian" if os_release else False


async def _check_debian_version(version: str, conn: SSHClientConnection) -> bool:
    "Check for Debian version"
    os_release = await _get_os_release(conn)
    return len(os_release) >= 2 and os_release[2] == version if os_release else False


check_is_debian_10 = partial(_check_debian_version, "10")
check_is_debian_11 = partial(_check_debian_version, "11")
check_is_debian_12 = partial(_check_debian_version, "12")
check_is_debian_13 = partial(_check_debian_version, "13")
check_is_debian_14 = partial(_check_debian_version, "14")
check_is_debian_15 = partial(_check_debian_version, "15")


@lru_cache
async def check_is_windows(conn: SSHClientConnection) -> bool:
    "Check for any Windows with Powershell"
    proc = await conn.run("[environment]::OSVersion")
    return proc.returncode == 0


@lru_cache
async def get_wmi_w32_os_caption(conn: SSHClientConnection) -> Optional[str]:
    "Get OS caption from WMI object"
    proc = await conn.run("(Get-WmiObject -class Win32_OperatingSystem).Caption")
    if proc.stdout:
        return str(proc.stdout)


async def _check_is_windows_caption_version(
    version: str, conn: SSHClientConnection
) -> bool:
    "Check for Windows version in caption"
    caption = await get_wmi_w32_os_caption(conn)
    return version in caption if caption else False


check_is_windows7 = partial(_check_is_windows_caption_version, "7")
check_is_windows8 = partial(_check_is_windows_caption_version, "8")
check_is_windows10 = partial(_check_is_windows_caption_version, "10")
check_is_windows11 = partial(_check_is_windows_caption_version, "11")
check_is_windows12 = partial(_check_is_windows_caption_version, "12")



#!/usr/bin/env python3

import asyncio
import logging
from abc import abstractmethod
from contextlib import asynccontextmanager
from datetime import timedelta
from pathlib import PurePath
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    ValuesView,
)

from asyncssh.connection import SSHClientConnection
from asyncssh.misc import (
    ChannelListenError,
    ChannelOpenError,
    CompressionError,
    ConnectionLost,
    KeyExchangeFailed,
    MACError,
    ProtocolError,
    ServiceNotAvailable,
)
from asyncssh.process import SSHClientProcess, SSHCompletedProcess
from asyncssh.sftp import (
    SFTPBadMessage,
    SFTPByteRangeLockConflict,
    SFTPByteRangeLockRefused,
    SFTPClient,
    SFTPConnectionLost,
    SFTPDeletePending,
    SFTPEOFError,
    SFTPFailure,
    SFTPInvalidHandle,
    SFTPLockConflict,
    SFTPNoConnection,
    SFTPNoMatchingByteRangeLock,
)
from typing_extensions import Protocol, Self, TypedDict, Unpack

SFTPRetryExc = (
    asyncio.TimeoutError,
    SFTPEOFError,
    SFTPFailure,
    SFTPBadMessage,
    SFTPNoConnection,
    SFTPConnectionLost,
    SFTPInvalidHandle,
    SFTPLockConflict,
    SFTPByteRangeLockConflict,
    SFTPByteRangeLockRefused,
    SFTPDeletePending,
    SFTPNoMatchingByteRangeLock,
)
SSHRetryExc = (
    OSError,
    asyncio.TimeoutError,
    CompressionError,
    ConnectionLost,
    KeyExchangeFailed,
    MACError,
    ProtocolError,
    ServiceNotAvailable,
    ChannelOpenError,
    ChannelListenError,
)
AllSSHRetryExc = SSHRetryExc + SFTPRetryExc


class PProcessInfo(Protocol):
    pid: int
    name: str
    command: str


class PEngine(Protocol):
    name: str
    deployable: Tuple
    platforms: Tuple[str, ...]
    check_pname: Optional[str]
    check_cmd: Optional[str]
    check_cmd_code: int
    sleep_interval: int


class PEngineRepository(Protocol):
    @abstractmethod
    def get_platform_packages(self) -> Sequence[str]:
        raise NotImplementedError

    @abstractmethod
    def filter_platforms(self, platforms: Sequence[str]) -> "PEngineRepository":
        raise NotImplementedError

    @abstractmethod
    def values(self) -> ValuesView[PEngine]:
        raise NotImplementedError


class PNode(Protocol):
    ip: str
    username: str


SSHCheck = Callable[[SSHClientConnection], Coroutine[Any, Any, bool]]
QuoteCallable = Callable[[str], str]


class RunCallable(Protocol):
    @abstractmethod
    def __call__(
        self,
        conn: SSHClientConnection,
        quote: QuoteCallable,
        command: str,
        *args,
        cwd: Optional[str] = None,
        **kwargs,
    ) -> Coroutine[Any, Any, SSHCompletedProcess]:
        pass


class RunBgCallable(Protocol):
    @abstractmethod
    def __call__(
        self,
        conn: SSHClientConnection,
        quote: QuoteCallable,
        command: str,
        *args,
        cwd: Optional[str] = None,
        **kwargs,
    ) -> Coroutine[Any, Any, SSHClientProcess]:
        pass


class OuterRunCallable(Protocol):
    @abstractmethod
    def __call__(
        self, *args, cwd: Optional[str] = None, **kwargs
    ) -> Coroutine[Any, Any, SSHCompletedProcess]:
        pass


GetCPUCoresCallable = Callable[[OuterRunCallable], Coroutine[Any, Any, int]]


class ListProcessesCallable(Protocol):
    @abstractmethod
    def __call__(
        self, conn: SSHClientConnection, query: Optional[str] = None
    ) -> AsyncGenerator[PProcessInfo, None]:
        pass


class PgrepCallable(Protocol):
    @abstractmethod
    def __call__(
        self,
        conn: SSHClientConnection,
        quote: QuoteCallable,
        pattern: Union[str, Pattern[str]],
        full=True,
    ) -> AsyncGenerator[PProcessInfo, None]:
        pass


class SetupNodeCallable(Protocol):
    @abstractmethod
    def __call__(
        self,
        conn: SSHClientConnection,
        run: OuterRunCallable,
        quote: QuoteCallable,
        engines: PEngineRepository,
        engines_dir: PurePath,
        log: Optional[logging.Logger] = None,
    ) -> Coroutine[Any, Any, None]:
        pass


class PRemoteMachineAdapter(Protocol):
    platform: str
    path: Type[PurePath]
    quote: QuoteCallable
    run: RunCallable
    run_bg: RunBgCallable
    checks: Sequence[SSHCheck]
    get_cpu_cores: GetCPUCoresCallable
    list_processes: ListProcessesCallable
    pgrep: PgrepCallable
    setup_node: SetupNodeCallable


class PRemoteMachineMetadata(Protocol):
    busy: Optional[bool]

    @abstractmethod
    def is_free_longer_than(self, delta: timedelta) -> bool:
        raise NotImplementedError


class PRemoteMachineCreateKwargsCommon(TypedDict):
    client_keys: Optional[Union[Sequence[bytes], Sequence[str]]]
    logger: Optional[logging.Logger]
    connect_timeout: Optional[float]
    data_dir: Optional[PurePath]
    engines_dir: Optional[PurePath]
    tasks_dir: Optional[PurePath]
    jump_host: Optional[str]
    jump_username: Optional[str]


class PRemoteMachineCreateKwargs(PRemoteMachineCreateKwargsCommon):
    host: str
    username: str


class PRemoteMachine(Protocol):
    "Remote SSH machine"

    meta: PRemoteMachineMetadata
    path: Type[PurePath]

    # supported platforms (like debian-10, debian, debian-like and linux)
    platforms: Sequence[str]

    data_dir: PurePath
    engines_dir: PurePath
    tasks_dir: PurePath

    jobs: Set[asyncio.Task]

    @abstractmethod
    def __le__(self, other: Self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __gt__(self, other: Self) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def create(
        cls, **kwargs: Unpack[PRemoteMachineCreateKwargs]
    ) -> "PRemoteMachine":
        """
        Async init of remote machine.
        If adapter is not set, then platform will be guessed.
        :raises PlatformGuessFailed: Not supported platform.
        :raises asyncssh.Error: An SSH error has occurred.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    @asynccontextmanager
    def create_ctx(cls, *args, **kwargs) -> AsyncGenerator["PRemoteMachine", None]:
        """
        Create async context.
        :raises asyncssh.Error: An SSH error has occurred.
        """
        raise NotImplementedError

    async def close(self) -> None:
        "Close connections and free resources"
        raise NotImplementedError

    @property
    @abstractmethod
    def hostname(self) -> str:
        raise NotImplementedError

    @abstractmethod
    @asynccontextmanager
    def sftp(self, **kwargs) -> AsyncGenerator[SFTPClient, None]:
        """
        Open SFTP connection
        :raises asyncssh.SFTPError: An SFTP error has occurred.
        """

    @abstractmethod
    def quote(self, s: str) -> str:
        "Platform-specific shell quoting"
        raise NotImplementedError

    @abstractmethod
    async def run(
        self, *args, cwd: Optional[str] = None, **kwargs
    ) -> SSHCompletedProcess:
        "Run process and wait for exit"
        raise NotImplementedError

    @abstractmethod
    async def run_bg(
        self, command: str, *args, cwd: Optional[str] = None, **kwargs
    ) -> SSHCompletedProcess:
        "Run process in background"

    @abstractmethod
    async def get_cpu_cores(self) -> int:
        "Get number of CPU cores"
        raise NotImplementedError

    @abstractmethod
    def list_processes(self) -> AsyncGenerator[PProcessInfo, None]:
        "Returns information about all running processes"
        raise NotImplementedError

    @abstractmethod
    def pgrep(
        self, pattern: Union[str, Pattern], full: bool = True
    ) -> AsyncGenerator[PProcessInfo, None]:
        """
        Returns information about running processes, that name matches a pattern.
        If `full`, check match against name or full cmd.
        """
        raise NotImplementedError

    @abstractmethod
    async def setup_node(self, engines: PEngineRepository):
        """
        Setup node for target engines.
        :raises NotImplemented: Not supported on platform.
        """
        raise NotImplementedError("Not implemented for this platform")

    @abstractmethod
    async def occupancy_check(self, engine: PEngine) -> bool:
        """
        Check node occupancy by task for target engine
        """
        raise NotImplementedError

    @abstractmethod
    async def start_occupancy_check(self, engine: PEngine) -> None:
        """
        Start occupancy checker for engine.
        """
        raise NotImplementedError



#!/usr/bin/env python3
"""Cloud configurations"""

from configparser import SectionProxy
from functools import partial
from typing import Optional, Sequence, Union

from attrs import define, field, fields, validators
from typing_extensions import Self

from .utils import _make_default_field, opt_str_val, warn_unknown_fields


def _check_az_user(_: "ConfigCloudAzure", __, value: str):
    if value == "root":
        raise ValueError("Root user is forbidden on Azure")


def _fmt_key(prefix: str, name: str):
    return f"{prefix}_{name}"


@define(frozen=True)
class AzureImageReference:
    """Azure's image reference"""

    publisher: str = field(default="Debian")
    offer: str = field(default="debian-11-daily")
    sku: str = field(default="11-backports-gen2")
    version: str = field(default="latest")

    @classmethod
    def from_urn(cls, urn: str) -> Self:
        "Create image reference from urn in forma `publisher:offer:sku:version'"
        parts = urn.split(":", maxsplit=4)
        if len(parts) < 4:
            raise ValueError(
                "`Image reference URN should be in format publisher:offer:sku:version"
            )

        return cls(*parts)


@define(frozen=True)
class ConfigCloudAzure:
    """Azure cloud configuration"""

    prefix = "az"
    tenant_id: str = field(validator=validators.instance_of(str))
    client_id: str = field(validator=validators.instance_of(str))
    client_secret: str = field(validator=validators.instance_of(str))
    subscription_id: str = field(validator=validators.instance_of(str))
    resource_group: str = _make_default_field("yascheduler-rg")
    location: str = _make_default_field("westeurope")
    vnet: str = _make_default_field("yascheduler-vnet")
    subnet: str = _make_default_field("yascheduler-subnet")
    nsg: str = _make_default_field("yascheduler-nsg")
    vm_image: AzureImageReference = _make_default_field(AzureImageReference())
    vm_size: str = _make_default_field("Standard_B1s")
    max_nodes: int = _make_default_field(10, extra_validators=[validators.ge(0)])
    username: str = _make_default_field(
        "yascheduler", extra_validators=[_check_az_user]
    )
    priority: int = _make_default_field(0)
    idle_tolerance: int = _make_default_field(300, extra_validators=[validators.ge(1)])
    jump_username: Optional[str] = field(default=None, validator=opt_str_val)
    jump_host: Optional[str] = field(default=None, validator=opt_str_val)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        exclude_names = ["prefix", "username", "jump_username", "vm_image", "vm_size"]
        include_names = ["user", "jump_user", "image", "size"]
        return [
            f"{cls.prefix}_{x}"
            for x in [f.name for f in fields(cls) if f.name not in exclude_names]
            + include_names
        ]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigCloudAzure":
        "Create config from config parser's section"

        fmt = partial(_fmt_key, cls.prefix)

        warn_unknown_fields(
            [
                *cls.get_valid_config_parser_fields(),
                *ConfigCloudHetzner.get_valid_config_parser_fields(),
                *ConfigCloudUpcloud.get_valid_config_parser_fields(),
            ],
            sec,
        )

        vm_image = sec.get(fmt("image"))
        image_ref = None
        if vm_image:
            image_ref = AzureImageReference.from_urn(vm_image)

        return cls(
            tenant_id=sec.get(fmt("tenant_id")),
            client_id=sec.get(fmt("client_id")),
            client_secret=sec.get(fmt("client_secret")),
            subscription_id=sec.get(fmt("subscription_id")),
            resource_group=sec.get(fmt("resource_group")),
            location=sec.get(fmt("location")),
            vnet=sec.get(fmt("vnet")),
            subnet=sec.get(fmt("subnet")),
            nsg=sec.get(fmt("nsg")),
            vm_image=image_ref or AzureImageReference(),
            vm_size=sec.get(fmt("size")),
            max_nodes=sec.getint(fmt("max_nodes")),
            username=sec.get(fmt("user")),
            priority=sec.getint(fmt("priority")),
            idle_tolerance=sec.getint(fmt("idle_tolerance")),
            jump_username=sec.get(fmt("jump_user"), None),
            jump_host=sec.get(fmt("jump_host"), None),
        )


@define(frozen=True)
class ConfigCloudHetzner:
    """Hetzner cloud configuration"""

    prefix = "hetzner"
    token: str = field(validator=validators.instance_of(str))
    max_nodes: int = _make_default_field(10, extra_validators=[validators.ge(0)])
    username: str = _make_default_field("root")
    priority: int = _make_default_field(0)
    server_type: str = _make_default_field("cx52")
    image_name: str = _make_default_field("debian-11")
    idle_tolerance: int = _make_default_field(120, extra_validators=[validators.ge(1)])
    jump_username: Optional[str] = field(default=None, validator=opt_str_val)
    jump_host: Optional[str] = field(default=None, validator=opt_str_val)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        exclude_names = ["prefix", "username", "jump_username"]
        include_names = ["user", "jump_user"]
        return [
            f"{cls.prefix}_{x}"
            for x in [f.name for f in fields(cls) if f.name not in exclude_names]
            + include_names
        ]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigCloudHetzner":
        "Create config from config parser's section"
        fmt = partial(_fmt_key, cls.prefix)

        warn_unknown_fields(
            [
                *ConfigCloudAzure.get_valid_config_parser_fields(),
                *cls.get_valid_config_parser_fields(),
                *ConfigCloudUpcloud.get_valid_config_parser_fields(),
            ],
            sec,
        )

        return cls(
            token=sec.get(fmt("token")),
            max_nodes=sec.getint(fmt("max_nodes")),
            username=sec.get(fmt("user")),
            server_type=sec.get(fmt("server_type")),
            image_name=sec.get(fmt("image_name")),
            priority=sec.getint(fmt("priority")),
            idle_tolerance=sec.getint(fmt("idle_tolerance")),
            jump_username=sec.get(fmt("jump_user"), None),
            jump_host=sec.get(fmt("jump_host"), None),
        )


@define(frozen=True)
class ConfigCloudUpcloud:
    """Upcloud cloud configuration"""

    prefix = "upcloud"
    login: str = field(validator=validators.instance_of(str))
    password: str = field(validator=validators.instance_of(str))
    max_nodes: int = _make_default_field(10, extra_validators=[validators.ge(0)])
    username: str = _make_default_field("root")
    priority: int = _make_default_field(0)
    idle_tolerance: int = _make_default_field(120, extra_validators=[validators.ge(1)])
    jump_username: Optional[str] = field(default=None, validator=opt_str_val)
    jump_host: Optional[str] = field(default=None, validator=opt_str_val)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        exclude_names = ["prefix", "username", "jump_username"]
        include_names = ["user", "jump_user"]
        return [
            f"{cls.prefix}_{x}"
            for x in [f.name for f in fields(cls) if f.name not in exclude_names]
            + include_names
        ]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigCloudUpcloud":
        "Create config from config parser's section"
        fmt = partial(_fmt_key, cls.prefix)

        warn_unknown_fields(
            [
                *ConfigCloudAzure.get_valid_config_parser_fields(),
                *ConfigCloudHetzner.get_valid_config_parser_fields(),
                *cls.get_valid_config_parser_fields(),
            ],
            sec,
        )

        return cls(
            login=sec.get(fmt("login")),
            password=sec.get(fmt("password")),
            max_nodes=sec.getint(fmt("max_nodes")),
            username=sec.get(fmt("user")),
            priority=sec.getint(fmt("priority")),
            idle_tolerance=sec.getint(fmt("idle_tolerance")),
            jump_username=sec.get(fmt("jump_user"), None),
            jump_host=sec.get(fmt("jump_host"), None),
        )


ConfigCloud = Union[ConfigCloudAzure, ConfigCloudHetzner, ConfigCloudUpcloud]



#!/usr/bin/env python3
"""Database configuration"""

from configparser import SectionProxy
from typing import Sequence

from attrs import define, fields

from .utils import _make_default_field, warn_unknown_fields


@define(frozen=True)
class ConfigDb:
    """Database configuration"""

    user: str = _make_default_field("yascheduler")
    password: str = _make_default_field("password")
    database: str = _make_default_field("database")
    host: str = _make_default_field("localhost")
    port: int = _make_default_field(5432)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        return [f.name for f in fields(cls)]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigDb":
        "Create config from config parser's section"
        warn_unknown_fields(cls.get_valid_config_parser_fields(), sec)
        return cls(
            sec.get("user"),
            sec.get("password"),
            sec.get("database"),
            sec.get("host"),
            sec.getint("port"),
        )



#!/usr/bin/env python
"""
Yascheduler systemd daemon
"""

if __name__ == "__main__":
    from yascheduler import LOG_FILE
    from yascheduler.utils import daemonize

    daemonize(log_file=LOG_FILE)



#!/usr/bin/env python3
"""Config helper utilities"""

import warnings
from configparser import SectionProxy
from typing import Optional, Sequence

from attrs import converters, field, validators

opt_str_val = validators.optional(validators.instance_of(str))


class ConfigWarning(Warning):
    "Warning about config"


def _make_default_field(default, extra_validators: Optional[Sequence] = None):
    return field(
        default=default,
        converter=converters.default_if_none(default=default),
        validator=[validators.instance_of(type(default)), *(extra_validators or [])],
    )


def warn_unknown_fields(known_fields: Sequence[str], sec: SectionProxy) -> None:
    unknown_fields = list(set(sec.keys()) - set(known_fields))
    if unknown_fields:
        warnings.warn(
            f"Config section {sec.name} unknown fields: {', '.join(unknown_fields)}",
            ConfigWarning,
            3,
        )



#!/usr/bin/env python3
"""Remote configuration"""

from configparser import SectionProxy
from pathlib import PurePath
from typing import Optional, Sequence

from attrs import define, field, fields

from .utils import _make_default_field, opt_str_val, warn_unknown_fields


@define(frozen=True)
class ConfigRemote:
    """Local configuration"""

    data_dir: PurePath = _make_default_field(PurePath("./data"))
    tasks_dir: PurePath = _make_default_field(PurePath("./data/tasks"))
    engines_dir: PurePath = _make_default_field(PurePath("./data/engines"))
    username: str = _make_default_field("root")
    jump_username: Optional[str] = field(default=None, validator=opt_str_val)
    jump_host: Optional[str] = field(default=None, validator=opt_str_val)

    @classmethod
    def get_valid_config_parser_fields(cls) -> Sequence[str]:
        "Returns a list of valid config keys"
        return [f.name for f in fields(cls)]

    @classmethod
    def from_config_parser_section(cls, sec: SectionProxy) -> "ConfigRemote":
        "Create config from config parser's section"
        warn_unknown_fields(cls.get_valid_config_parser_fields(), sec)
        data_dir = PurePath(sec.get("data_dir", "./data"))
        return cls(
            data_dir=data_dir,
            engines_dir=PurePath(sec.get("engines_dir", str(data_dir / "engines"))),
            tasks_dir=PurePath(sec.get("tasks_dir", str(data_dir / "tasks"))),
            username=sec.get("user"),
            jump_username=sec.get("jump_user", None),
            jump_host=sec.get("jump_host", None),
        )



"""Cloud adapters"""

import asyncio
from functools import lru_cache
from typing import Sequence, Tuple

from attrs import define, field

from .az import az_create_node, az_delete_node
from .hetzner import hetzner_create_node, hetzner_delete_node
from .protocols import (
    CreateNodeCallable,
    DeleteNodeCallable,
    PCloudAdapter,
    SupportedPlatformChecker,
    TConfigCloud_contra,
)
from .upcloud import upcload_delete_node, upcloud_create_node


def can_debian_buster(platform: str) -> bool:
    "Platform is compatible with Debian Buster"
    return platform in ["debian-10", "debian", "debian-like", "linux"]


def can_debian_bullseye(platform: str) -> bool:
    "Platform is compatible with Debian Bullseye"
    return platform in ["debian-11", "debian", "debian-like", "linux"]


def can_win10(platform: str) -> bool:
    "Platform is compatible with Windows 10"
    return platform in ["windows-10", "windows"]


def can_win11(platform: str) -> bool:
    "Platform is compatible with Windows 11"
    return platform in ["windows-11", "windows"]


@define(frozen=True)
class CloudAdapter(PCloudAdapter[TConfigCloud_contra]):
    """Cloud adapter"""

    name: str = field()
    supported_platform_checks: Tuple[SupportedPlatformChecker] = field()
    create_node: CreateNodeCallable[TConfigCloud_contra] = field()
    create_node_conn_timeout: int = field()
    create_node_timeout: int = field()
    delete_node: DeleteNodeCallable[TConfigCloud_contra] = field()
    op_limit: int = field(default=1)

    @classmethod
    def create(
        cls,
        name: str,
        supported_platform_checks: Sequence[SupportedPlatformChecker],
        create_node: CreateNodeCallable[TConfigCloud_contra],
        delete_node: DeleteNodeCallable[TConfigCloud_contra],
        create_node_conn_timeout: int = 10,
        create_node_timeout: int = 300,
        op_limit: int = 1,
    ):
        return cls(
            name=name,
            supported_platform_checks=tuple(supported_platform_checks),
            create_node=create_node,
            create_node_conn_timeout=create_node_conn_timeout,
            create_node_timeout=create_node_timeout,
            delete_node=delete_node,
            op_limit=op_limit,
        )

    @lru_cache  # noqa: B019
    def get_op_semaphore(self) -> asyncio.Semaphore:
        return asyncio.Semaphore(self.op_limit)


azure_adapter = CloudAdapter.create(
    name="az",
    supported_platform_checks=[can_debian_bullseye, can_win11],
    create_node=az_create_node,
    delete_node=az_delete_node,
    op_limit=5,
)
hetzner_adapter = CloudAdapter.create(
    name="hetzner",
    supported_platform_checks=[can_debian_buster],
    create_node=hetzner_create_node,
    delete_node=hetzner_delete_node,
    op_limit=5,
)
upcloud_adapter = CloudAdapter.create(
    name="upcloud",
    supported_platform_checks=[can_debian_buster],
    create_node=upcloud_create_node,
    delete_node=upcload_delete_node,
    op_limit=1,
)



"""
Console scripts for yascheduler
"""

import argparse
import asyncio
import logging
import os
import signal
from pathlib import Path
from typing import Any, Mapping, Sequence

from pg8000 import ProgrammingError

from .client import Yascheduler, to_sync
from .config import Config
from .db import DB, TaskModel, TaskStatus
from .remote_machine import RemoteMachine
from .scheduler import Scheduler, get_logger
from .variables import CONFIG_FILE


@to_sync
async def submit():
    parser = argparse.ArgumentParser(
        description="Submit task to yascheduler via AiiDA script"
    )
    parser.add_argument("script")

    args = parser.parse_args()
    script_file = Path(args.script)
    if not script_file.exists():
        raise ValueError("Script parameter is not a file name")

    logging.captureWarnings(True)
    log = logging.getLogger()
    log.setLevel(logging.WARN)
    yac = Yascheduler(logger=log)

    script_params = {}
    with script_file.open("r") as f:
        for line in f.readlines():
            try:
                k, v = line.split("=")
                script_params[k.strip()] = v.strip()
            except ValueError:
                pass

    label = script_params.get("LABEL", "AiiDA job")
    metadata: Mapping[str, Any] = {
        "local_folder": os.getcwd()
    }  # NB AiiDA chdirs to repo, but if not?
    if not script_params.get("ENGINE"):
        raise ValueError("Script has not defined an engine")

    engine = yac.config.engines.get(script_params["ENGINE"])
    if not engine:
        raise ValueError("Engine %s is not supported" % script_params["ENGINE"])

    for input_file in engine.input_files:
        try:
            metadata[input_file] = Path(
                metadata["local_folder"], input_file
            ).read_text()
        except Exception as err:
            raise ValueError(
                "Script was not supplied with the required input file"
            ) from err

    webhook_onsubmit = False
    if "PARENT" in script_params and yac.config.local.webhook_url:
        metadata["webhook_url"] = yac.config.local.webhook_url
        metadata["webhook_custom_params"] = {"parent": script_params["PARENT"]}
        webhook_onsubmit = True

    task_id = await yac.queue_submit_task_async(
        label,
        metadata,
        engine.name,
        webhook_onsubmit=webhook_onsubmit,
    )

    # this should be received by AiiDA
    print(str(task_id))


@to_sync
async def check_status():  # noqa: C901
    parser = argparse.ArgumentParser(description="Show status of tasks")
    parser.add_argument("-j", "--jobs", required=False, default=None, nargs="*")
    parser.add_argument(
        "-v", "--view", required=False, default=None, nargs="?", type=bool, const=True
    )
    parser.add_argument(
        "-o",
        "--convergence",
        required=False,
        default=None,
        nargs="?",
        type=bool,
        const=True,
        help="needs -v option",
    )
    parser.add_argument(
        "-i", "--info", required=False, default=None, nargs="?", type=bool, const=True
    )
    # parser.add_argument(
    #     "-k", "--kill", required=False, default=None, nargs="?", type=bool, const=True
    # )

    args = parser.parse_args()
    config = Config.from_config_parser(CONFIG_FILE)
    db = await DB.create(config.db)

    local_parsing_ready, local_calc_snippet = False, False

    tasks: Sequence[TaskModel] = []
    if args.jobs:
        tasks = await db.get_tasks_by_jobs(jobs=args.jobs)
    else:
        tasks = await db.get_tasks_by_status(
            statuses=(TaskStatus.RUNNING, TaskStatus.TO_DO)
        )

    if args.convergence:
        local_parsing_ready = True

    if args.view:
        for task in await db.get_tasks_with_cloud_by_id_status(
            ids=list(map(lambda x: x.task_id, tasks)), status=TaskStatus.RUNNING
        ):
            ssh_user = None
            for c in config.clouds:
                ssh_user = c.username
            ssh_user = ssh_user or config.remote.username
            print(
                "." * 50
                + "ID%s %s at %s@%s:%s:%s"
                % (
                    task.task_id,
                    task.label,
                    ssh_user,
                    task.ip,
                    task.cloud or "",
                    task.metadata.get("remote_folder", ""),
                )
            )
            machine = await RemoteMachine.create(
                host=task.ip,
                username=ssh_user,
                client_keys=config.local.get_private_keys(),
            )
            r_output = machine.path(task.metadata.get("remote_folder")) / "OUTPUT"
            result = await machine.run(f"tail -n15 {machine.quote(str(r_output))}")
            if result.returncode:
                print("OUTDATED TASK, SKIPPING")
            else:
                print(result.stdout)

            if local_parsing_ready:
                local_calc_snippet = Path(
                    config.local.data_dir, "local_calc_snippet.tmp"
                )
                try:
                    r_output = (
                        machine.path(task.metadata.get("remote_folder")) / "OUTPUT"
                    )
                    async with machine.sftp() as sftp:
                        await sftp.get([str(r_output)], local_calc_snippet)
                except OSError:
                    continue

                # pylint: disable=import-error
                from numpy import nan

                # pylint: disable=import-error
                from pycrystal import CRYSTOUT, CRYSTOUT_Error

                try:
                    calc = CRYSTOUT(local_calc_snippet)
                except CRYSTOUT_Error as err:
                    print(err)
                    continue
                output_lines = ""
                if calc.info["convergence"]:
                    output_lines += str(calc.info["convergence"]) + "\n"
                if calc.info["optgeom"]:
                    for n in range(len(calc.info["optgeom"])):
                        try:
                            ncycles = calc.info["ncycles"][n]
                        except IndexError:
                            ncycles = "^"
                        output_lines += (
                            "{:8f}".format(calc.info["optgeom"][n][0] or nan)
                            + "  "
                            + "{:8f}".format(calc.info["optgeom"][n][1] or nan)
                            + "  "
                            + "{:8f}".format(calc.info["optgeom"][n][2] or nan)
                            + "  "
                            + "{:8f}".format(calc.info["optgeom"][n][3] or nan)
                            + "  "
                            + "E={:12f}".format(calc.info["optgeom"][n][4] or nan)
                            + " eV"
                            + "  "
                            + "(%s)" % ncycles
                            + "\n"
                        )
                print(output_lines)

    # elif args.kill:
    #    if not args.jobs:
    #        print('NO JOBS GIVEN')
    #        return
    #    yac.cursor.execute(
    #        'SELECT ip FROM yascheduler_tasks WHERE status=%s AND task_id IN (%s);' % (
    #        yac.STATUS_RUNNING, ', '.join([str(task['task_id']) for task in tasks])
    #    ))
    #    for row in yac.cursor.fetchall():
    #        ssh_conn = SSH_Connection(host=row[0], user=config.get('remote', 'user'),
    #            connect_kwargs=ssh_custom_key)
    #        try:
    #            result = ssh_conn.run('pkill %s' % yac.RUNNING_MARKER, hide=True)
    #        except: pass

    elif args.info:
        for task in tasks:
            print(
                "task_id={}\tstatus={}\tlabel={}\tip={}".format(
                    task.task_id, task.status.name, task.label, task.ip or "-"
                )
            )

    else:
        for task in tasks:
            print(f"{task.task_id}   {task.status.name}")

    await db.close()

    if local_calc_snippet and os.path.exists(local_calc_snippet):
        os.unlink(local_calc_snippet)


@to_sync
async def init():
    # service initialization
    install_path = Path(__file__).parent
    # check for systemd (exit status is 0 if there is a process)
    has_systemd = not os.system("pidof systemd")
    if has_systemd:
        # NB. will be absent in *service --status-all*
        _init_systemd(install_path)
    else:
        _init_sysv(install_path)
    await _init_db(install_path)


def _init_systemd(install_path: Path):
    print("Installing systemd service")
    # create unit file in /lib/systemd/system
    src_unit_file = install_path / "data/yascheduler.service"
    unit_file = Path("/lib/systemd/system/yascheduler.service")
    if not unit_file.is_file():
        if not os.access(unit_file, os.W_OK):
            print("Error: cannot write to %s" % unit_file)
            return
        daemon_file = install_path / "daemon_systemd.py"
        systemd_script = src_unit_file.read_text("utf-8").replace(
            "%YASCHEDULER_DAEMON_FILE%", str(daemon_file)
        )
        unit_file.write_text(systemd_script, "utf-8")


def _init_sysv(install_path: Path):
    print("Installing SysV service")

    # create sysv script in /etc/init.d
    src_startup_file = install_path / "data/yascheduler.sh"
    startup_file = Path("/etc/init.d/yascheduler")
    if not startup_file.is_file():
        if not os.access(startup_file, os.W_OK):
            print("Error: cannot write to %s" % startup_file)
            return

        daemon_file = install_path / "daemon_sysv.py"
        sysv_script = src_startup_file.read_text("utf-8").replace(
            "%YASCHEDULER_DAEMON_FILE%", str(daemon_file)
        )
        startup_file.write_text(sysv_script, "utf-8")
        # make script executable
        os.chmod(startup_file, 0o755)


async def _init_db(install_path: Path):
    # database initialization
    config = Config.from_config_parser(CONFIG_FILE)
    db = await DB.create(config.db, automigrate=False)
    schema = (install_path / "data" / "schema.sql").read_text()
    try:
        await db.run(schema)
        await db.commit()
        await db.close()
    except ProgrammingError as e:
        if "already exists" in str(e.args[0]):
            print("Database already initialized!")
        raise


@to_sync
async def show_nodes():
    config = Config.from_config_parser(CONFIG_FILE)
    db = await DB.create(config.db)

    tasks = await db.get_tasks_by_status(statuses=[TaskStatus.RUNNING])
    nodes = await db.get_all_nodes()
    for node in nodes:
        tmpl = (
            "ip={ip} ncpus={ncpus} enabled={enabled} "
            "occupied_by={occ} (task_id={tid}) {cloud}"
        )
        node_tasks = list(filter(lambda x: x.ip == node.ip, tasks))
        node_label = "-"
        task_id = "-"
        for x in node_tasks:
            node_label = x.label
            task_id = x.task_id
        msg = tmpl.format(
            ip=node.ip,
            ncpus=node.ncpus or "MAX",
            enabled=node.enabled,
            occ=node_label,
            tid=task_id,
            cloud=node.cloud or "",
        )
        print(msg)


@to_sync
async def manage_node():
    parser = argparse.ArgumentParser(description="Add nodes to yascheduler daemon")
    parser.add_argument("host", help="[user@]IP[~ncpus]")
    parser.add_argument(
        "--skip-setup",
        required=False,
        default=False,
        nargs="?",
        type=bool,
        const=True,
        help="Skip node setup",
    )
    parser.add_argument(
        "--remove-soft",
        required=False,
        default=None,
        nargs="?",
        type=bool,
        const=True,
        help="Remove IP delayed",
    )
    parser.add_argument(
        "--remove-hard",
        required=False,
        default=None,
        nargs="?",
        type=bool,
        const=True,
        help="Remove IP immediate",
    )

    args = parser.parse_args()
    config = Config.from_config_parser(CONFIG_FILE)
    db = await DB.create(config.db)

    ncpus = None
    username = config.remote.username
    if "@" in args.host:
        username, args.host = args.host.split("@")
    if "~" in args.host:
        args.host, ncpus = args.host.split("~")
        ncpus = int(ncpus)

    already_there = await db.has_node(args.host)
    if already_there and not args.remove_hard and not args.remove_soft:
        print(f"Host already in DB: {args.host}")
        return False

    if not already_there and (args.remove_hard or args.remove_soft):
        print(f"Host NOT in DB: {args.host}")
        return False

    if args.remove_hard:
        task_ids = await db.get_task_ids_by_ip_and_status(args.host, TaskStatus.RUNNING)
        for task_id in task_ids:
            await db.update_task_status(task_id, TaskStatus.DONE)
            print(
                "An associated task {} at {} is now marked done!".format(
                    task_id, args.host
                )
            )

        await db.remove_node(args.host)
        await db.commit()
        await db.close()
        print(f"Removed host from yascheduler: {args.host}")
        return True

    elif args.remove_soft:
        task_ids = await db.get_task_ids_by_ip_and_status(args.host, TaskStatus.RUNNING)
        if task_ids:
            print("A task associated, prevent from assigning the new tasks")
            await db.disable_node(args.host)
            print(f"Prevented from assigning the new tasks: {args.host}")
        else:
            print("No tasks associated, remove node immediately")
            await db.remove_node(args.host)
            print(f"Removed host from yascheduler: {args.host}")
        await db.commit()
        await db.close()
        return True

    machine = await RemoteMachine.create(
        host=args.host,
        username=username,
        client_keys=config.local.get_private_keys(),
        engines_dir=config.remote.engines_dir,
    )

    if not args.skip_setup:
        print("Setup host...")
        await machine.setup_node(config.engines)

    await db.add_node(ip_addr=args.host, username=username, ncpus=ncpus, enabled=True)
    await db.commit()
    await db.close()

    print(f"Added host to yascheduler: {args.host}")


def daemonize(log_file=None):
    parser = argparse.ArgumentParser(description="Start yascheduler daemon")
    parser.add_argument(
        "-l",
        "--log-level",
        default="INFO",
        help="set log level",
        choices=logging._levelToName.values(),
    )
    args = parser.parse_args()

    logger = get_logger(log_file, level=logging._nameToLevel[args.log_level])

    async def on_signal(
        y: Scheduler, shield: Sequence[asyncio.Task], sig: signal.Signals
    ):
        signame = signal.strsignal(sig)
        logger.info(f"Received signal {signame}")
        if sig in [signal.SIGTERM, signal.SIGINT]:
            await y.stop()
            shielded = [*shield, asyncio.current_task()]
            tasks = [t for t in asyncio.all_tasks() if t not in shielded]
            logger.info(f"Cancelling {len(tasks)} outstanding tasks")
            [task.cancel() for task in tasks]
            await asyncio.gather(*tasks, return_exceptions=True)
            # Wait 250 ms for the underlying SSL connections to close
            await asyncio.sleep(0.25)
            logger.info("Done")

    async def run():
        yac = await Scheduler.create(log=logger)

        loop = asyncio.get_running_loop()
        current_task = asyncio.current_task()

        shielded = [current_task] if current_task else []
        for sig in [signal.SIGTERM, signal.SIGINT]:

            def handler():
                task = on_signal(yac, shielded, sig)  # noqa: B023
                return asyncio.create_task(task)

            loop.add_signal_handler(sig, handler)

        await yac.start()

    to_sync(run)()



#!/usr/bin/env python
"""
SystemV Daemon functions
"""

import argparse
import os

import daemon
from daemon import pidfile

from yascheduler import LOG_FILE, PID_FILE
from yascheduler.utils import daemonize


def start_daemon(pid_file, log_file):
    """Launch daemon in its context as per
    https://stackoverflow.com/questions/13106221/"""
    with daemon.DaemonContext(
        working_directory=os.path.dirname(__file__),
        umask=0o002,
        pidfile=pidfile.TimeoutPIDLockFile(pid_file),
    ) as _:
        daemonize(log_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Yascheduler daemon")
    parser.add_argument("-p", "--pid-file", default=PID_FILE)
    parser.add_argument("-l", "--log-file", default=LOG_FILE)

    args = parser.parse_args()

    start_daemon(pid_file=args.pid_file, log_file=args.log_file)
